/**
 * @file irq.c
 *
 * This file contains code to handle Lexra IRQs. Majority of the code was 
 * generated from generic Linux interrupt handling functions. Additional 
 * code added to implement the Lexra vectored interrupts and to provide 
 * support for up to 16 PCI interrupts.
 */

#include <linux/errno.h>
#include <linux/init.h>
#include <linux/kernel_stat.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/timex.h>
#include <linux/malloc.h>
#include <linux/random.h>

#include <asm/bitops.h>
#include <asm/bootinfo.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/mipsregs.h>
#include <asm/system.h>

#ifdef ISHONI_ENGINE
#include <asm/ibe/ibeirq.h>
#else
#include <asm/lexra/lxirq.h>
#endif
#include <asm/lexra/lx_defs.h>

#ifdef CONFIG_REMOTE_DEBUG
extern void breakpoint(void);
#endif
#ifdef ISHONI_ENGINE  /* NAZIM */
extern int lx_irq_setup();
/* I think we should have a different routine to register CPU ISR.
 *
 * int registerCpuIsr(int irqNumber, void (*handler), int intPriority);
 * intPriority: is used as index to this table.
 *           where zero is the highest priority and 7 the lowest.
 * irqNumber: Actual interrupt number, requires for masking.
 * handler  : Interrupt handler.
 *
 * Check this with Deepak.
 */
 
/* Index zero we have given to NetPIC Handler. */ 
int	cpuHandleIndexAvalilable = 1; 
extern void putString(char *string,int len);
#endif

/**
 * External function defined in int-handler.S. This is an assembly language
 * routine used to mask the Lexra vectored interrupts using Lexra specific
 * instructions.
 */
extern inline void mask_lx_interrupt(unsigned int mask);

/**
 * External function defined in int-handler.S. This is an assembly language
 * routine used to unmask the Lexra vectored interrupts using Lexra specific
 * instructions.
 */
extern inline void unmask_lx_interrupt(unsigned int mask);

/**
 * Spurious interrupt counter referenced in entry.S.
 */
unsigned long spurious_count = 0;

extern int pci_devices_count; 

/**
 * Internal function to mask an interrupt. Will mask any of the three
 * types of Lexra interrupts: standard MIPS, Lexra vectored or PCI.
 * @param irq_nr The interrupt number to be masked.
 */
inline void mask_irq(unsigned int irq_nr)
{		
#ifdef ISHONI_ENGINE /* NAZIM */

        if(irq_nr >= NR_INTS)
        {
                printk("-------- Wrong irq number %d passed in mask_irq.-------\n",irq_nr);
                return;
        }

        if (ibe_interrupt[irq_nr].iemask) { /* This is an ASIC interrupt    */
               RESET_PIC_IEN(~(ibe_interrupt[irq_nr].iemask));
        } else               /* This is a cpu interrupt        */
             set_cp0_status(ST0_IM, read_32bit_cp0_register(CP0_STATUS) & ~ibe_interrupt[irq_nr].cpu_mask);

#else 
    u32 cp0;
	/* This is a lexra vectored interrupt    */
	if (lx_interrupt[irq_nr].LX_mask != 0) {
		mask_lx_interrupt(~lx_interrupt[irq_nr].LX_mask);
	}
	else if(lx_interrupt[irq_nr].cpu_mask != 0) { 	
		/* This is a cpu interrupt        */
		set_cp0_status(ST0_IM, read_32bit_cp0_register(CP0_STATUS) & 
				~lx_interrupt[irq_nr].cpu_mask);
	}
#endif
}

/**
 * Internal function to unmask an interrupt. Will unmask any of the three
 * types of Lexra interrupts: standard MIPS, Lexra vectored or PCI.
 * @param irq_nr The interrupt number to be unmasked.
 */
static inline void unmask_irq(unsigned int irq_nr)
{	
#ifdef ISHONI_ENGINE /* NAZIM */

        if(irq_nr >= NR_INTS)
        {
                printk("-------- Wrong irq number %d passed in unmask_irq.----------\n",irq_nr);
                return;
        }

        if (ibe_interrupt[irq_nr].iemask) {   /* This is an ASIC interrupt    */
               SET_PIC_IEN(ibe_interrupt[irq_nr].iemask);
    } else                      /* This is a cpu interrupt        */
    set_cp0_status(ST0_IM, read_32bit_cp0_register(CP0_STATUS) | ibe_interrupt[irq_nr].cpu_mask);

#else  
    u32 cp0;
	/* This is a Lexra vectored interrupt    */
	if (lx_interrupt[irq_nr].LX_mask != 0) {
		unmask_lx_interrupt(lx_interrupt[irq_nr].LX_mask);
	}
	else if (lx_interrupt[irq_nr].cpu_mask != 0){	
		/* This is a cpu interrupt        */
		set_cp0_status(ST0_IM, read_32bit_cp0_register(CP0_STATUS) | 
				lx_interrupt[irq_nr].cpu_mask);
	}
#endif
}

/**
 * Externally visable funtion used to disable an interrupt. Function handles 
 * save and restore operations before and after calling the internal 
 * mask_irq() function.  
 * @param irq_nr The interrupt number to be disabled.
 */
void disable_irq(unsigned int irq_nr)
{
	unsigned long flags;

	save_and_cli(flags);
	mask_irq(irq_nr);
	restore_flags(flags);
}

/**
 * Externally visable funtion used to enable an interrupt. Function handles 
 * save and restore operations before and after calling the internal 
 * unmask_irq() function.  
 * @param irq_nr The interrupt number to be disabled.
 */
void enable_irq(unsigned int irq_nr)
{
	unsigned long flags;

	save_and_cli(flags);
	unmask_irq(irq_nr);
	restore_flags(flags);
}

/**
 * The array of irqaction structures to store all information relating to 
 * the available interrupt numbers. Each structure contains the information 
 * required to route the interrupt to the appropriate handler in the do_IRQ() 
 * function.
 */
static struct irqaction *irq_action[32] =
{
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
};

/**
 * Function to retrieve a interrupt usage report. The function walks the 
 * irqaction array and records the interrupt information in a tabular form to 
 * the buffer buf.
 * @param buf Pointer to a charater buffer to receive the results.
 * @return The length of the buffer returned.
 */
int get_irq_list(char *buf)
{
	int i, len = 0;
	struct irqaction *action;

	for (i = 0; i < 32; i++) {
		action = irq_action[i];
		if (!action)
		    continue;
		len += sprintf(buf + len, "%2d: %8d %c %s",
			       i, kstat.irqs[0][i],
			       (action->flags & SA_INTERRUPT) ? '+' : ' ',
			       action->name);
		for (action = action->next; action; action = action->next) {
		    len += sprintf(buf + len, ",%s %s",
			       (action->flags & SA_INTERRUPT) ? " +" : "",
			       action->name);
		}
		len += sprintf(buf + len, "\n");
	}
	return len;
}

/**
 * The function do_IRQ() handles all Lexra IRQ's that have been installed.
 * If the interrupt is installed with the SA_INTERRUPT flag it uses the
 * full signal-handling return and runs with other interrupts enabled.
 * All relatively slow IRQ's should use this format: notably the keyboard/timer
 * routines.
 * @param irq the interrupt number
 * @param regs pointer to the stack passed from the assemble interupt handler
 */
asmlinkage void do_IRQ(int irq, struct pt_regs *regs)
{
	struct irqaction *action;
	int do_random, cpu;

	cpu = smp_processor_id();
	irq_enter(cpu, irq);

	kstat.irqs[cpu][irq]++;
	mask_irq(irq);
	action = *(irq + irq_action);

	if (action) {
//		if (!(action->flags & SA_INTERRUPT)) __sti();

		do_random = 0;
		do {
			do_random |= action->flags;
			action->handler(irq, action->dev_id, regs);
			action = action->next;
		} while (action);
		if (do_random & SA_SAMPLE_RANDOM)
			add_interrupt_randomness(irq);
	}
	__cli();
	unmask_irq(irq);
	irq_exit(cpu, irq);

	/* unmasking and bottom half handling is done magically for us. */
}

/**
 * Lexra specific interrupt setup routine. Idea is to put all interrupts
 * in a single table and differenciate them just by number.
 * @param irq The interrupt number (0 - 31).
 * @param new A pointer to the irqaction structure for interrupt being 
 * initialized.
 * The irqaction structure is defined in /include/asm/irq.h.
 * @return Function returns a 0 if successful, an error code if unsuccessful.
 */
int setup_lx_irq(int irq, struct irqaction *new)
{
	int shared = 0;
	struct irqaction *old, **p;
	unsigned long flags;

	p = irq_action + irq;
	if ((old = *p) != NULL) {
		/* Can't share interrupts unless both agree to */
		if (!(old->flags & new->flags & SA_SHIRQ)) return -EBUSY;

		/* Can't share interrupts unless both are same type */
		if ((old->flags ^ new->flags) & SA_INTERRUPT) return -EBUSY;

		/**
		 * Add new interrupt to end of irq queue.
		 */
		do {
			p = &old->next;
			old = *p;
		} while (old);
		shared = 1;
	}
	if (new->flags & SA_SAMPLE_RANDOM)
		rand_initialize_irq(irq);

#ifndef ISHONI_ENGINE  /* NAZIM */ 
    if (irq >= 16)				/* enable this pci interrupt */
        lx_interrupt[irq].pci_mask[1] = 1;
#endif

#ifdef ISHONI_ENGINE     /* caught the bug. */
    /* Set up host interrupt.*/
	if(irq >= MAX_CPU_INTS ){
        if(irq == 9){
            unmask_irq(irq+1);
            unmask_irq(irq+2);
         }
        /* its a netpic interrupt.*/ 
    	ibe_interrupt[irq].cpu_mask = IE_IRQ4;
    	ibe_interrupt[irq].iemask = (1 << PIC_INT(irq));
   		netpic_irq_nr[PIC_INT(irq)] = irq;
	}else{
        /* Actually going by the architecture all the CPU interrupts
         * should not come to do_IRQ functions at all.
         * They are maintained in cpu_ivec_table[] and exceuted directly
         * from the ibe_handle_int() just like nvPicHandler();
         */  
    }
#endif
	save_and_cli(flags);
	*p = new;
#ifndef ISHONI_ENGINE
	if(irq != PCI_ABCD) {
#endif
		if (!shared) unmask_irq(irq); 
#ifndef ISHONI_ENGINE
	}
	else
		if(--pci_devices_count <= 0 ) unmask_irq(irq); 
#endif
	restore_flags(flags);
	return 0;
}

/**
 * Function provided for use by a device driver to request an interrupt. The
 * funtion uses the paramters supplied to fill in the related interrupt 
 * irqaction structure and subsequently register the interrupt if the 
 * requested interrupt number is available.
 * @param irq The requested interrupt number (0-31).
 * @param handler Pointer to the interrupt service routine.
 * @param irqflags List of flags descibing the interrupt.
 * @param devname Pointer to a character string containing the device name.
 * @param dev_id Void pointer to a device id parameter.
 * @return Returns 0 if successful, error code otherwise.
 */
int request_irq(unsigned int irq,
		void (*handler) (int, void *, struct pt_regs *),
		unsigned long irqflags,
		const char *devname,
		void *dev_id)
{
	int retval;
	struct irqaction *action;

	if (irq >= 32)
		return -EINVAL;
	if (!handler)
		return -EINVAL;

#ifdef ISHONI_ENGINE
	if( irq < MAX_CPU_INTS ){
		/* Handle CPU interrupts here.
     	 * we are not supposed to call them through do_IRQ function.
     	 * They are maintaned in cpu_ivec_table[] and called directly 
     	 * from this table in the ibe_handle_int().
     	 */  
    	cpu_ivec_tbl[cpuHandleIndexAvalilable] = (unsigned int)handler;
    	cpuHandleIndexAvalilable++;
 		return 0;
    }
#endif
	action = (struct irqaction *) 
		kmalloc(sizeof(struct irqaction), GFP_KERNEL);
	if (!action)
		return -ENOMEM;

	action->handler = handler;
	action->flags = irqflags;
	action->mask = 0;
	action->name = devname;
	action->next = NULL;
	action->dev_id = dev_id;

	retval = setup_lx_irq(irq, action);

	if (retval)
	kfree(action);
	return retval;
}

/**
 * Function provided to free an interrupt when no longer in use.
 * @param irq The interrupt to be freed.
 * @param dev_id Pointer to the device id.
 */
void free_irq(unsigned int irq, void *dev_id)
{
	struct irqaction *action, **p;
	unsigned long flags;

	if (irq >= 32) {
		printk(KERN_DEBUG "Trying to free IRQ%d\n", irq);
		return;
	}
	for (p = irq + irq_action; (action = *p) != NULL; p = &action->next) {
		if (action->dev_id != dev_id)
		    continue;

		/* Found it - now free it */
		save_and_cli(flags);
		*p = action->next;
		if (!irq[irq_action]) 
			mask_irq(irq);
		restore_flags(flags);
		kfree(action);
		return;
	}
}

/**
 * Unimplemented interrupt probing routine.
 */
unsigned long probe_irq_on(void)
{
	/* TODO */
	return 0;
}

/**
 * Unimplemented interrupt probing routine.
 */
int probe_irq_off(unsigned long irqs)
{
	/* TODO */
	return 0;
}

/**
 * Initial setup of interrupts. Called by start_kernel(). Only
 * function is to  call the function irq_setup().
 */
void __init init_IRQ(void)
{
	lx_irq_setup();

#ifdef CONFIG_REMOTE_DEBUG
	/* If local serial I/O used for debug port, enter kgdb at once */
#ifndef ISHONI_ENGINE
	puts("Waiting for kgdb to connect...");
#else
	putString("Waiting for kgdb to connect...",30);
#endif
	set_debug_traps();
	breakpoint(); 
#endif
}

