/*
 * cache.c: MMU and cache operations for the LX4xxx/5xxx
 *
 *
 */

#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <asm/mipsregs.h>
#include <asm/page.h>
#include <asm/pgtable.h>
#include <asm/system.h>
#include <asm/mmu_context.h>
#include <asm/isadep.h>
#include <asm/io.h>
#include <asm/wbflush.h>
#include <asm/cpu.h>

/**
 * This structure contains processor related information.
 * The tlbsize is stored in this structure.
 */
extern struct mips_cpu mips_cpu;

#undef DEBUG_TLB

/**
 * The icace and dcache size variables are currently set to hard values,
 * as there is no way to size the Lexra caches through software.
 */
int icache_size, dcache_size;

void lx_clear_page(void *);
static void lx_copy_page(void *, void *);

/**
 * This function is used to flush a region of the instruction cache. Because
 * Lexra processors do not accomodate instruction cache flushing using 
 * traditional MIPS methods, this function invalidates the entire instruction 
 * cache. This is accomplished using the Lexra specific IINVAL bit of the 
 * coprocessor 0 cache control register.
 * @param start The start of the memory range to flush. Ignored.
 * @param size The size in bytes of the memory range to flush. Ignored.
 */
static void __lx_flush_icache_range(unsigned long start,
				    unsigned long size)
{
	unsigned int controlReg;
	unsigned long flags;

	save_and_cli(flags);

	controlReg = read_32bit_cp0_register(CP0_CCTL);

	write_32bit_cp0_register(CP0_CCTL, (controlReg & ~CCTL_IINVAL));
	write_32bit_cp0_register(CP0_CCTL, (controlReg | CCTL_IINVAL));

	/* delay to allow cache to be flushed */
	__asm__ __volatile__(".set\tnoreorder\n\t"
			     "nop\n\t"
			     "nop\n\t"
			     "nop\n\t"
			     "nop\n\t"
			     "nop\n\t"
			     "nop\n\t"
			     "nop\n\t"
			     "nop\n\t" "nop\n\t" ".set\treorder\n\t");

	write_32bit_cp0_register(CP0_CCTL, controlReg);
	restore_flags(flags);
}

/**
 * This function is used to flush a region of the data cache. To accomplish this
 * with a Lexra processor, we read from the corresponding uncached space in
 * KSEG1.
 * @param start The start of the memory range to flush.
 * @param size The size in bytes of the memory range to flush.
 */
static void __lx_flush_dcache_range(unsigned long start,
				    unsigned long size)
{
	int index;
	register volatile int *ptr;
	register int dummy;

	unsigned long flags;
	size = (size + 8) >> 2;	/* bytes to words rounded to nearest word */

	/* Don't have to do anything to KSEG1 addresses */
	if (start == (unsigned long) KSEG1ADDR(start))
		return;

	save_and_cli(flags);

	ptr = (volatile int *) KSEG1ADDR(start & 0xfffffffc);
	for (index = 0; index < size; index++)
		dummy += *ptr++;

	restore_flags(flags);

}


/**
 * This function is used to flush the entire instruction cache. This is
 * done by calling lx_flush_cache_range() and specifying the whole of KSEG0 as
 * the range.
 */
static inline void lx_flush_cache_all(void)
{
	__lx_flush_icache_range(KSEG0, KSEG0 + icache_size);
}

/*
 * This function is used to flush a portion of the instruction cache specified
 * by a mm_struct. For the Lexra processor we can only flush the entire cache 
 * so this is done by calling lx_flush_cache_range() and specifying the whole 
 * of KSEG0 as the range.
 * @param mm A mm_struct pointer.
 */
static void lx_flush_cache_mm(struct mm_struct *mm)
{
	if (mm->context != 0) {
		__lx_flush_icache_range(KSEG0, KSEG0 + icache_size);
	}
}

/**
 * This function is used to flush a portion of the instruction cache specified
 * by a mm_struct. For the Lexra processor we can only flush the entire cache 
 * so this is done by calling lx_flush_cache_range() and specifying the whole 
 * of KSEG0 as the range.
 * @param mm A mm_struct pointer.
 * @param start Start of memory region to flush.
 * @param end End of memory region to flush.
 */
static void lx_flush_cache_range(struct mm_struct *mm,
				 unsigned long start, unsigned long end)
{
	struct vm_area_struct *vma;

	if (mm->context == 0)
		return;

	start &= PAGE_MASK;

	vma = find_vma(mm, start);

	if (!vma)
		return;
	/* we are always going to just flush the entire cache for now */

	__lx_flush_icache_range(KSEG0, KSEG0 + icache_size);

}

/**
 * This function is used to flush a portion of the instruction cache specified
 * by a vm_area_struct. For the Lexra processor we can only flush the entire 
 * cache so this is done by calling lx_flush_cache_range() and specifying the 
 * whole of KSEG0 as the range.
 * @param vma A vm_area_struct pointer.
 * @param page The page to flush.
 */
static void lx_flush_cache_page(struct vm_area_struct *vma,
				unsigned long page)
{
	struct mm_struct *mm = vma->vm_mm;

	if (mm->context == 0)
		return;

	__lx_flush_icache_range(KSEG0, KSEG0 + icache_size);
}

/**
 * Lexra has a write through cache, therefore, this is an empty
 * function in the Lexra Linux implementation.
 */
static void lx_flush_page_to_ram(struct page *page)
{
	/* Nothing to be done */
}


/**
 * This function is used to flush a portion of the instruction cache specified
 * by a vm_area_struct. For the Lexra processor we can only flush the entire 
 * cache so this is done by calling lx_flush_icache_range() and specifying the 
 * whole of KSEG0 as the range.
 * @param vma A vm_area_struct pointer.
 * @param page Pointer to a page structure.
 */
static void lx_flush_icache_page(struct vm_area_struct *vma,
				 struct page *page)
{
	struct mm_struct *mm = vma->vm_mm;
	unsigned long physpage;

	if (mm->context == 0)
		return;

	if (!(vma->vm_flags & VM_EXEC))
		return;

	physpage = (unsigned long) page_address(page);
	if (physpage)
		__lx_flush_icache_range(KSEG0, KSEG0 + icache_size);
}

/**
 * Traditional R3000 processors flush the first three
 * words ot the cache starting at addr. We must flush
 * the entire cache.
 * @param addr Variable specifying the beginning of the  flush.
 */
static void lx_flush_cache_sigtramp(unsigned long addr)
{
	/**
	 * Traditional R3000 processors flush the first three
	 * words ot the cache starting at addr. We must flush
	 * the entire cache.
	 */
	__lx_flush_icache_range(addr, icache_size);
}

/**
 * This function is used to flush and invalidate the memory used for DMA 
 * accesses.  Because Lexra uses a write through cache, it is only necessary 
 * to flush the specified range of the data cache.
 * @param start The start of the memory range used in the DMA transfer.
 * @param size The size in bytes of the DMA transfer.
 */
static void __lx_dma_cache_wback_inv(unsigned long start,
				     unsigned long size)
{
	__lx_flush_dcache_range(start, size);
}


/**
 * This function initializes all of the Lexra memory management functions.
 * This involves mapping the kernel cache management function pointers
 * to the Lexra specific functions, as well as setting the icache and
 * dcache size variables.
 */
void __init ld_mmu_lx(void)
{
	_clear_page = lx_clear_page;
	_copy_page = lx_copy_page;

	/*
	 * These variables are unused on this architecture.
	 */
	icache_size = 1024 * 4;
	dcache_size = 1024 * 2;

	_flush_cache_all = lx_flush_cache_all;
	___flush_cache_all = lx_flush_cache_all;
	_flush_cache_mm = lx_flush_cache_mm;
	_flush_cache_range = lx_flush_cache_range;
	_flush_cache_page = lx_flush_cache_page;
	_flush_cache_sigtramp = lx_flush_cache_sigtramp;
	_flush_page_to_ram = lx_flush_page_to_ram;
	_flush_icache_page = lx_flush_icache_page;

	_flush_icache_range = __lx_flush_icache_range;
	// _flush_dcache_range = __lx_flush_dcache_range;

	_dma_cache_wback_inv = __lx_dma_cache_wback_inv;

	_flush_cache_all();
	flush_tlb_all();
}

/* From r2300.c */

/**
 * Assebly language function to clear a page of memory.
 * @param page Pointer to the page.
 */
void lx_clear_page(void *page)
{
	__asm__ __volatile__(".set\tnoreorder\n\t"
			     ".set\tnoat\n\t"
			     "addiu\t$1,%0,%2\n"
			     "1:\tsw\t$0,(%0)\n\t"
			     "sw\t$0,4(%0)\n\t"
			     "sw\t$0,8(%0)\n\t"
			     "sw\t$0,12(%0)\n\t"
			     "addiu\t%0,32\n\t"
			     "sw\t$0,-16(%0)\n\t"
			     "sw\t$0,-12(%0)\n\t"
			     "sw\t$0,-8(%0)\n\t"
			     "bne\t$1,%0,1b\n\t"
			     "sw\t$0,-4(%0)\n\t"
			     ".set\tat\n\t"
			     ".set\treorder":"=r"(page):"0"(page),
			     "I"(PAGE_SIZE):"$1", "memory");
}


/**
 * This function is used to copy a page (4096 bytes) from the location
 * pointed to by 'from' to the location pointed to by 'to'.
 * @param to Location to copy to.
 * @param from Location to copy from.
 */
static void lx_copy_page(void *to, void *from)
{
	unsigned long dummy1, dummy2;
	unsigned long reg1, reg2, reg3, reg4;

	/* This function copies a 4 kb page from addr 'from' to the addr
	 * pointed to by 'to'.
	 */
	__asm__ __volatile__(".set\tnoreorder\n\t"
			     ".set\tnoat\n\t"
			     "addiu\t$1,%0,%8\n"
			     "1:\tlw\t%2,(%1)\n\t"
			     "lw\t%3,4(%1)\n\t"
			     "lw\t%4,8(%1)\n\t"
			     "lw\t%5,12(%1)\n\t"
			     "sw\t%2,(%0)\n\t"
			     "sw\t%3,4(%0)\n\t"
			     "sw\t%4,8(%0)\n\t"
			     "sw\t%5,12(%0)\n\t"
			     "lw\t%2,16(%1)\n\t"
			     "lw\t%3,20(%1)\n\t"
			     "lw\t%4,24(%1)\n\t"
			     "lw\t%5,28(%1)\n\t"
			     "sw\t%2,16(%0)\n\t"
			     "sw\t%3,20(%0)\n\t"
			     "sw\t%4,24(%0)\n\t"
			     "sw\t%5,28(%0)\n\t"
			     "addiu\t%0,64\n\t"
			     "addiu\t%1,64\n\t"
			     "lw\t%2,-32(%1)\n\t"
			     "lw\t%3,-28(%1)\n\t"
			     "lw\t%4,-24(%1)\n\t"
			     "lw\t%5,-20(%1)\n\t"
			     "sw\t%2,-32(%0)\n\t"
			     "sw\t%3,-28(%0)\n\t"
			     "sw\t%4,-24(%0)\n\t"
			     "sw\t%5,-20(%0)\n\t"
			     "lw\t%2,-16(%1)\n\t"
			     "lw\t%3,-12(%1)\n\t"
			     "lw\t%4,-8(%1)\n\t"
			     "lw\t%5,-4(%1)\n\t"
			     "sw\t%2,-16(%0)\n\t"
			     "sw\t%3,-12(%0)\n\t"
			     "sw\t%4,-8(%0)\n\t"
			     "bne\t$1,%0,1b\n\t"
			     "sw\t%5,-4(%0)\n\t"
			     ".set\tat\n\t"
			     ".set\treorder":"=r"(dummy1), "=r"(dummy2),
			     "=&r"(reg1), "=&r"(reg2), "=&r"(reg3),
			     "=&r"(reg4):"0"(to), "1"(from),
			     "I"(PAGE_SIZE));
}

/**
 * This function is used to display the current state of the processor
 * by displaying all of the standard MIPS registers. Called whenever a
 * kernel opps occurs, an unhandled memory exception.
 * @param regs Pointer to the pt_regs structure containing the state of
 * the processors registers.
 */
void show_regs(struct pt_regs *regs)
{
	/*
	 * Saved main processor registers
	 */
	printk("$0 : %08x %08lx %08lx %08lx %08lx %08lx %08lx %08lx\n",
	       0, (unsigned long) regs->regs[1],
	       (unsigned long) regs->regs[2],
	       (unsigned long) regs->regs[3],
	       (unsigned long) regs->regs[4],
	       (unsigned long) regs->regs[5],
	       (unsigned long) regs->regs[6],
	       (unsigned long) regs->regs[7]);
	printk("$8 : %08lx %08lx %08lx %08lx %08lx %08lx %08lx %08lx\n",
	       (unsigned long) regs->regs[8],
	       (unsigned long) regs->regs[9],
	       (unsigned long) regs->regs[10],
	       (unsigned long) regs->regs[11],
	       (unsigned long) regs->regs[12],
	       (unsigned long) regs->regs[13],
	       (unsigned long) regs->regs[14],
	       (unsigned long) regs->regs[15]);
	printk("$16: %08lx %08lx %08lx %08lx %08lx %08lx %08lx %08lx\n",
	       (unsigned long) regs->regs[16],
	       (unsigned long) regs->regs[17],
	       (unsigned long) regs->regs[18],
	       (unsigned long) regs->regs[19],
	       (unsigned long) regs->regs[20],
	       (unsigned long) regs->regs[21],
	       (unsigned long) regs->regs[22],
	       (unsigned long) regs->regs[23]);
	printk
	    ("$24: %08lx %08lx                   %08lx %08lx %08lx %08lx\n",
	     (unsigned long) regs->regs[24],
	     (unsigned long) regs->regs[25],
	     (unsigned long) regs->regs[28],
	     (unsigned long) regs->regs[29],
	     (unsigned long) regs->regs[30],
	     (unsigned long) regs->regs[31]);

	/*
	 * Saved cp0 registers
	 */
	printk("epc  : %08lx\nStatus: %08x\nCause : %08x\n",
	       (unsigned long) regs->cp0_epc,
	       (unsigned int) regs->cp0_status,
	       (unsigned int) regs->cp0_cause);
}

/**
 * This function is used to flush the entire TLB.
 */
void flush_tlb_all(void)
{
	unsigned long flags;
	unsigned long old_ctx;
	int entry;

#ifdef DEBUG_TLB
	printk("[tlball]\n");
#endif

	save_and_cli(flags);
	old_ctx = (get_entryhi() & 0xfc0);
	write_32bit_cp0_register(CP0_ENTRYLO0, 0);
	for (entry = 0; entry < mips_cpu.tlbsize; entry++) {
		write_32bit_cp0_register(CP0_INDEX, entry << 8);
		write_32bit_cp0_register(CP0_ENTRYHI,
					 ((entry | 0x80000) << 12));
		__asm__ __volatile__("tlbwi");
	}
	set_entryhi(old_ctx);
	restore_flags(flags);
}

/**
 * This function is used to flush the region of the TLB associated
 * with the memory reference by the mm_struct pointer provided.
 * @param mm The mm_struct varible describing the affected memory region.
 */
void flush_tlb_mm(struct mm_struct *mm)
{
	if (mm->context != 0) {
		unsigned long flags;

#ifdef DEBUG_TLB
		printk("[tlbmm<%d>]\n", mm->context);
#endif
		save_and_cli(flags);
		get_new_mmu_context(mm, asid_cache);
		if (mm == current->active_mm)
			set_entryhi(mm->context & 0xfc0);
		restore_flags(flags);
	}
}

/**
 * This function is used to flush TLB entries related to the memory
 * region referenced by the mm_struct variable in the range defined
 * by the start and end parameters.
 * @param mm The mm_struct varible describing the affected memory region.
 * @param start The start of the memory range.
 * @param end The end of the memory range.
 */
void flush_tlb_range(struct mm_struct *mm, unsigned long start,
		     unsigned long end)
{
	if (mm->context != 0) {
		unsigned long flags;
		int size;

#ifdef DEBUG_TLB
		printk("[tlbrange<%02x,%08lx,%08lx>]\n",
		       (mm->context & 0xfc0), start, end);
#endif
		save_and_cli(flags);
		size = (end - start + (PAGE_SIZE - 1)) >> PAGE_SHIFT;
		if (size <= mips_cpu.tlbsize) {
			int oldpid = (get_entryhi() & 0xfc0);
			int newpid = (mm->context & 0xfc0);

			start &= PAGE_MASK;
			end += (PAGE_SIZE - 1);
			end &= PAGE_MASK;
			while (start < end) {
				int idx;

				set_entryhi(start | newpid);
				start += PAGE_SIZE;
				tlb_probe();
				idx = get_index();
				set_entrylo0(0);
				set_entryhi(KSEG0);
				if (idx < 0)
					continue;
				tlb_write_indexed();
			}
			set_entryhi(oldpid);
		} else {
			get_new_mmu_context(mm, asid_cache);
			if (mm == current->active_mm)
				set_entryhi(mm->context & 0xfc0);
		}
		restore_flags(flags);
	}
}

/**
 * This function is used to flush TLB entries related to the virtual memory
 * page referenced by the vm_area_struct variable and page parameter.
 * @param vma The vm_area_struct varible describing the affected memory region.
 * @param page The page of memory.
 */
void flush_tlb_page(struct vm_area_struct *vma, unsigned long page)
{
	if (vma->vm_mm->context != 0) {
		unsigned long flags;
		int oldpid, newpid, idx;

#ifdef DEBUG_TLB
		printk("[tlbpage<%d,%08lx>]\n", vma->vm_mm->context, page);
#endif
		newpid = (vma->vm_mm->context & 0xfc0);
		page &= PAGE_MASK;
		save_and_cli(flags);
		oldpid = (get_entryhi() & 0xfc0);
		set_entryhi(page | newpid);
		tlb_probe();
		idx = get_index();
		set_entrylo0(0);
		set_entryhi(KSEG0);
		if (idx < 0)
			goto finish;
		tlb_write_indexed();

	      finish:
		set_entryhi(oldpid);
		restore_flags(flags);
	}
}

/**
 * Initialize new page directory with pointers to invalid page table
 * entries.
 * @param page The page to initialize.
 */
void pgd_init(unsigned long page)
{
	unsigned long dummy1, dummy2;

	/*
	 * The plain and boring version for the R3000.  No cache flushing
	 * stuff is implemented since the R3000 has physical caches.
	 */
	__asm__ __volatile__(".set\tnoreorder\n"
			     "1:\tsw\t%2,(%0)\n\t"
			     "sw\t%2,4(%0)\n\t"
			     "sw\t%2,8(%0)\n\t"
			     "sw\t%2,12(%0)\n\t"
			     "sw\t%2,16(%0)\n\t"
			     "sw\t%2,20(%0)\n\t"
			     "sw\t%2,24(%0)\n\t"
			     "sw\t%2,28(%0)\n\t"
			     "subu\t%1,1\n\t"
			     "bnez\t%1,1b\n\t"
			     "addiu\t%0,32\n\t"
			     ".set\treorder":"=r"(dummy1), "=r"(dummy2)
			     :"r"((unsigned long) invalid_pte_table),
			     "0"(page),
			     "1"(PAGE_SIZE / (sizeof(pmd_t) * 8)));
}

/**
 * Updates the page table.
 * @param vma Pointer to a vm_area_struct stucture.
 * @param address The address involved.
 * @param pte A page table entry descriptor structure    .
 */
void update_mmu_cache(struct vm_area_struct *vma,
		      unsigned long address, pte_t pte)
{
	unsigned long flags;
	pgd_t *pgdp;
	pmd_t *pmdp;
	pte_t *ptep;
	int idx, pid;

	/*
	 * Handle debugger faulting in for debugee.
	 */
	if (current->active_mm != vma->vm_mm)
		return;

	pid = get_entryhi() & 0xfc0;

#ifdef DEBUG_TLB
	if ((pid != (vma->vm_mm->context & 0xfc0))
	    || (vma->vm_mm->context == 0)) {
		printk
		    ("update_mmu_cache: Wheee, bogus tlbpid mmpid=%d tlbpid=%d\n",
		     (int) (vma->vm_mm->context & 0xfc0), pid);
	}
#endif

	save_and_cli(flags);
	address &= PAGE_MASK;
	set_entryhi(address | (pid));
	pgdp = pgd_offset(vma->vm_mm, address);
	tlb_probe();
	pmdp = pmd_offset(pgdp, address);
	idx = get_index();
	ptep = pte_offset(pmdp, address);
	set_entrylo0(pte_val(*ptep));
	set_entryhi(address | (pid));
	if (idx < 0) {
		tlb_write_random();
#if 0
		printk("[MISS]");
#endif
	} else {
		tlb_write_indexed();
#if 0
		printk("[HIT]");
#endif
	}
#if 0
	if (!strcmp(current->comm, "args")) {
		printk("<");
		for (idx = 0; idx < mips_cpu.tlbsize; idx++) {
			set_index(idx);
			tlb_read();
			address = get_entryhi();
			if ((address & 0xfc0) != 0)
				printk("[%08lx]", address);
		}
		printk(">\n");
	}
#endif
	set_entryhi(pid);
	restore_flags(flags);
}

/**
 * Unused stub required to compile arch/mips/kernel/setup.c
 */
unsigned long r3k_cache_size(void)
{
	return 0;
}
#ifdef ISHONI_ENGINE
void ishoni_patch_tlb(unsigned long address)
{
        unsigned long flags = 0;
        int pid;
        int idx;

        pid = get_entryhi() & 0xfc0;
        address &= PAGE_MASK;
        set_entryhi(address | (pid));
        tlb_probe();
        idx = get_index();
        if(idx > 0)
                panic("ishoni_patch_tlb: TLB invalidation entry for %lux already presentin TL
B\n",address);
        set_entryhi(address | (pid));
        set_entrylo0(address | (_PAGE_GLOBAL | _PAGE_VALID | _PAGE_DIRTY));
        tlb_write_random();
        restore_flags(flags);
}
#endif  /* ifdef ISHONI_ENGINE */

