/*
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * PCI support for NEC Vr4122.
 *
 * Copyright (C) 2000 Mike McDonald
 */
#include <linux/config.h>
#include <linux/delay.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/types.h>

#include <asm/byteorder.h>
#include <asm/io.h>
#include <asm/vr4122/vr4122.h>

#ifdef CONFIG_PCI

#define mkaddr(dev, reg)				\
do {							\
    if (dev->bus->number == 0) {			\
	       if (PCI_SLOT(dev->devfn) < 11 ||		\
		   PCI_SLOT(dev->devfn) >= 32 ||	\
		   PCI_FUNC(dev->devfn) >= 8 ||		\
		   reg >= 256)				\
		   return 0xFFFFFFFF;			\
	*(volatile u32 *)VR4122_PCICONFAREG =		\
            (reg & 0xFC) |				\
	    ( PCI_FUNC(dev->devfn) << 8) |		\
	    (1 << PCI_SLOT(dev->devfn));		\
    } else {						\
	       if (dev->bus->number >= 256 ||		\
		   PCI_SLOT(dev->devfn) >= 32 ||	\
		   PCI_FUNC(dev->devfn) >=8 ||		\
		   reg >= 256)				\
		   return 0xFFFFFFFF;			\
	*(volatile u32 *)VR4122_PCICONFAREG = 1 |	\
	    (reg & 0xFC) |				\
	    (dev->devfn << 8) |				\
	    (dev->bus->number << 16);			\
    }							\
} while (0);

/*
 * We can't address 8 and 16 bit words directly.  Instead we have to
 * read/write a 32bit word and mask/modify the data we actually want.
 */
int vr4122_pcibios_read_config_byte(struct pci_dev *dev,
                                    int where, unsigned char *val)
{
	u32 res;

	mkaddr(dev, where);
	res = readl(VR4122_PCICONFDREG);
	*val = (le32_to_cpu(res) >> ((where & 3) << 3)) & 0xff;

	return PCIBIOS_SUCCESSFUL;
}

int vr4122_pcibios_read_config_word(struct pci_dev *dev,
                                    int where, unsigned short *val)
{
	u32 res;

	if (where & 1)
		return PCIBIOS_BAD_REGISTER_NUMBER;
	mkaddr(dev, where);
	res = readl(VR4122_PCICONFDREG);
	*val = (le32_to_cpu(res) >> ((where & 3) << 3)) & 0xffff;

	return PCIBIOS_SUCCESSFUL;
}

int vr4122_pcibios_read_config_dword(struct pci_dev *dev,
                                     int where, unsigned int *val)
{
	u32 res;

	if (where & 3)
		return PCIBIOS_BAD_REGISTER_NUMBER;
	mkaddr(dev, where);
	res = readl(VR4122_PCICONFDREG);
	*val = le32_to_cpu(res);

	return PCIBIOS_SUCCESSFUL;
}

int vr4122_pcibios_write_config_byte(struct pci_dev *dev,
                                     int where, unsigned char val)
{
	mkaddr(dev, where);
	writeb(val, VR4122_PCICONFDREG + (where & 3));

	return PCIBIOS_SUCCESSFUL;
}

int vr4122_pcibios_write_config_word(struct pci_dev *dev,
                                     int where, unsigned short val)
{
	if (where & 1)
		return PCIBIOS_BAD_REGISTER_NUMBER;
	mkaddr(dev, where);
	writew(le16_to_cpu(val), VR4122_PCICONFDREG + (where & 3));

	return PCIBIOS_SUCCESSFUL;
}

int vr4122_pcibios_write_config_dword(struct pci_dev *dev,
                                      int where, unsigned int val)
{
	if (where & 3)
		return PCIBIOS_BAD_REGISTER_NUMBER;
	mkaddr(dev, where);
	writel(le32_to_cpu(val), VR4122_PCICONFDREG);

	return PCIBIOS_SUCCESSFUL;
}

struct pci_ops vr4122_pci_ops = {
	vr4122_pcibios_read_config_byte,
	vr4122_pcibios_read_config_word,
	vr4122_pcibios_read_config_dword,
	vr4122_pcibios_write_config_byte,
	vr4122_pcibios_write_config_word,
	vr4122_pcibios_write_config_dword
};

void __init vr4122_pcibios_init(void)
{
	u32 dummy;

	/* Enable PCI clock */
	vr4122_clock_supply(VR4122_CMUCLKMSK_MSKPCIU);
	udelay(100);

	/* Set master memory & I/O windows */
	*(volatile u32 *)VR4122_PCIMMAW1REG = 0x100f9010;
	*(volatile u32 *)VR4122_PCIMMAW2REG = 0x140fd014;
	*(volatile u32 *)VR4122_PCIMIOAWREG = 0x160fd000;

	/* Set target memory windows */
	*(volatile u32 *)VR4122_PCITAW1REG = 0x00081000;
	*(volatile u32 *)VR4122_PCIMBA1REG = 0x00000000;
	*(volatile u32 *)VR4122_PCITAW2REG = 0x00000000;
	*(volatile u32 *)VR4122_PCIMAILBAREG = 0x00000000;

	/* Clear bus error */
	dummy = *(volatile u32 *)VR4122_BUSERRADREG;

	*(volatile u32 *)VR4122_PCICLKSELREG = 0x00000000;      /* PCIckl = VRCKL/2 */
	*(volatile u32 *)VR4122_PCITRDYVREG = 100;
	*(volatile u32 *)VR4122_PCICACHELSREG = 0x00008004;

	udelay(100);

	*(volatile u32 *)VR4122_PCIENREG = 0x00000004;
	*(volatile u32 *)VR4122_PCICOMMANDREG = 0x00000147;

	mdelay(3);
}

#endif /* CONFIG_PCI */
