/*
 * Copyright 1999, 2000 Zereau Ltd.
 * Author: Alex Holden <alex@linuxhacker.org>
 */

#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/init.h>
#include <linux/socket.h>
#include <linux/net.h>
#include <linux/inet.h>
#include <linux/in.h>
#include <linux/uio.h>
#include <linux/string.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/blkdev.h>
#include <linux/timer.h>

#include <asm/uaccess.h>

#include "initrd_tftp.h"

#ifndef ISHONI_ENGINE
static char filename[TFTP_FILELEN] __initdata = CONFIG_BLK_DEV_INITRD_TFTP_FILE;
static char serverip[TFTP_SERVERNAMELEN] __initdata =
					CONFIG_BLK_DEV_INITRD_TFTP_SERVER;
#else
#ifdef LINUX_BOOT_LOADER
#define CONFIG_BLK_DEV_INITRD_TFTP_SERVER "0.0.0.0"
#define CONFIG_BLK_DEV_INITRD_TFTP_FILE "not-needed"
extern int getHostIPAddr(void);
extern int ipnumtostr(char *, int);
extern int getFilePathName(char *,int);
extern int getTftpFilePathName(char *);
static char serverip[TFTP_SERVERNAMELEN];
static char filename[TFTP_FILELEN];
#endif
#endif /* ISHONI_ENGINE */


/* 
 * Create a UDP socket and bind to it.
 */
static int __init bindsocket(struct initrd_tftp_state *tftpstate)
{
	int err;
	struct sockaddr_in sin;

	err = sock_create(PF_INET, SOCK_DGRAM, IPPROTO_UDP, &tftpstate->sock);
	if(err < 0) {
		printk(KERN_ERR "tftp: couldn\'t create socket\n");
		return 1;
	}
	
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = INADDR_ANY;
	sin.sin_port = 0;
	err = tftpstate->sock->ops->bind(tftpstate->sock,
					(struct sockaddr*) &sin, sizeof(sin));
        if(err < 0) {
		printk(KERN_ERR "tftp: couldn\'t bind to a port\n");
		sock_release(tftpstate->sock);
		return 1;
	}

	return 0;
}

/*
 * A very minimal strerror() replacement. If you get an "Unknown error"
 * under valid circumstances, please add a line for that condition...
 */
static void __init handleerror(struct initrd_tftp_state *tftpstate)
{
	switch(tftpstate->errno) {
		case -ECONNREFUSED:
			printk(KERN_INFO "tftp: Connection refused\n");
			break;
		case -EHOSTUNREACH:
			printk(KERN_INFO "tftp: Host unreachable\n");
			break;
		case -ENETUNREACH:
			printk(KERN_INFO "tftp: Network unreachable\n");
		default:
			printk(KERN_INFO "tftp: Unknown error %d\n",
							tftpstate->errno);
			break;
	}
}

/*
 * Send a signal to ourselves to force the sock_recvmsg() to return.
 */
static void __init timeout(unsigned long pid)
{
	kill_proc((pid_t) pid, SIGALRM, 1);
}

/*
 * Send out the packet in tftpstate->outbuf, which is tftpstate->outbuflen
 * bytes long, via the tftpstate->socket socket to port tftpstate->serverport
 * on host tftpstate->serverip. Returns 0 on success or -errno on failure.
 */
static int __init sendpacket(struct initrd_tftp_state *tftpstate)
{
	int len;
	struct sockaddr_in sin;
	struct msghdr msg;
	struct iovec iov;
	mm_segment_t old_fs;

	msg.msg_flags = 0;
	iov.iov_base = tftpstate->outbuf;
	iov.iov_len = tftpstate->outbuflen;
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = tftpstate->serverip;
	sin.sin_port = tftpstate->serverport;
	msg.msg_name = (struct sockaddr *) &sin;
	msg.msg_namelen = sizeof(sin);
	msg.msg_control = NULL;
	msg.msg_controllen = 0;

	old_fs = get_fs();
	set_fs(get_ds());
	len = sock_sendmsg(tftpstate->sock, &msg, tftpstate->outbuflen);
	set_fs(old_fs);
	if(len < 0) {
		tftpstate->errno = len;
		handleerror(tftpstate);	
		return len;
	}

	return 0;
}

/*
 * Send out a RRQ (read request) packet for tftpstate->file to
 * tftpstate->servername.
 */
static int __init sendrequest(struct initrd_tftp_state *tftpstate)
{
	int len;
	struct tftp_rqhdr *packet = (struct tftp_rqhdr*)tftpstate->outbuf;

	len = strnlen(serverip, TFTP_SERVERNAMELEN);
	if(len == TFTP_SERVERNAMELEN || len == 0) {
		printk(KERN_INFO "tftp: Bad server IP\n");
		return -1;
	}

	len = strnlen(filename, TFTP_FILELEN) + 1;
	if(len == (TFTP_FILELEN + 1) || len == 1) {
		printk(KERN_INFO "tftp: Bad filename\n");
		return -1;
	}

#ifndef ISHONI_ENGINE
	printk(KERN_INFO "tftp: Retrieving %s from %s (attempt %d)\n",
						filename, serverip,
						(TFTP_RRQ_MAX -
						tftpstate->retries));
#else
	printk("tftp: Retrieving %s from %s (attempt %d)\n",
						filename, serverip,
						(TFTP_RRQ_MAX -
						tftpstate->retries));
#endif

	packet->opcode = htons(TFTP_RRQ);
	memcpy(packet->data, filename, len);
	memcpy(&packet->data[len], TFTP_XFER, sizeof(TFTP_XFER));
	tftpstate->outbuflen = len + sizeof(TFTP_XFER) + 2;
	tftpstate->serverip = in_aton(serverip);

	return sendpacket(tftpstate);
}

/*
 * Attempt to receive a packet from tftpstate->socket and place it in
 * tftpstate->inbuf along with the length of the packet in tftpstate->inbuflen.
 * Packets which are too large are dropped with a warning. If this is the
 * first packet we receive from the server, we remember what port No. the
 * reply came from in tftpstate->serverport. Returned is the value we got
 * from sock_recvmsg(), which is either the number of bytes received or -errno.
 */
static int __init recvpacket(struct initrd_tftp_state *tftpstate)
{
	int len;
	mm_segment_t old_fs;
	struct sockaddr_in sin;
	struct msghdr msg;
	struct iovec iov;

	msg.msg_flags = 0;
	iov.iov_base = tftpstate->inbuf;
	iov.iov_len = sizeof(tftpstate->inbuf);
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_name = (struct sockaddr *) &sin;
	msg.msg_namelen = sizeof(sin);
	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	
	old_fs = get_fs();
	set_fs(get_ds());
	len = sock_recvmsg(tftpstate->sock, &msg, sizeof(tftpstate->inbuf), 0);
	set_fs(old_fs);

	if(msg.msg_flags & MSG_TRUNC) {
		printk(KERN_INFO "tftp: Dropping oversized packet\n");
		return -EAGAIN;
	}

	tftpstate->lastport = sin.sin_port;
	tftpstate->inbuflen = len;

	return len;
}

/*
 * Attempt to send an acknowledgement packet for the last data packet received.
 */
static int __init sendack(struct initrd_tftp_state *tftpstate)
{
	struct tftp_ackhdr *packet = (struct tftp_ackhdr*)tftpstate->outbuf;

	packet->opcode = htons(TFTP_ACK);
	packet->seqno = htons(tftpstate->seqno);
	tftpstate->outbuflen = 4;

	return sendpacket(tftpstate);
}

/*
 * Attempt to send an error packet to the server.
 * Be careful that outbuf is big enough if you add/change any strings.
 */
static int __init senderr(struct initrd_tftp_state *tftpstate)
{
	struct tftp_errhdr *packet = (struct tftp_errhdr*)tftpstate->outbuf;

	packet->opcode = htons(TFTP_ERROR);
	packet->errno = htons(tftpstate->errno);

	switch(tftpstate->errno) {
		case TFTP_EBADOP:
			sprintf(packet->msg, "Unknown TFTP packet type");
			tftpstate->outbuflen = 29;
			break;
		case TFTP_EBADID:
			sprintf(packet->msg, "Unknown source port");
			tftpstate->outbuflen = 24;
			break;
#if 0
		case TFTP_ENOTFOUND:
			sprintf(packet->msg, "File not found");
			tftpstate->outbuflen = 19;
			break;
		case TFTP_EACCESS:
			sprintf(packet->msg, "Access denied");
			tftpstate->outbuflen = 18;
			break;
		case TFTP_ENOSPACE:
			sprintf(packet->msg, "Out of space");
			tftpstate->outbuflen = 17;
			break;
		case TFTP_EEXISTS:
			sprintf(packet->msg, "File already exists");
			tftpstate->outbuflen = 24;
			break;
		case TFTP_ENOUSER:
			sprintf(packet->msg, "No such user");
			tftpstate->outbuflen = 17;
			break;
#endif
		default:
			packet->msg[0] = 0;
			tftpstate->outbuflen = 5;
	}

	return sendpacket(tftpstate);
}

/*
 * Print a human readable message when an error packet is received.
 * The rfc isn't clear on whether, if you don't want to send an error string,
 * you should just not send it or whether you should send a zero length string,
 * so we handle both cases. In our senderr(), we send a zero length string.
 */
static void __init handleerrorpacket(struct initrd_tftp_state *tftpstate)
{
	struct tftp_errhdr *packet = (struct tftp_errhdr*)tftpstate->inbuf;

	tftpstate->errno = ntohs(packet->errno);

	printk(KERN_INFO "tftp: Received ");
	switch(tftpstate->errno) {
		case TFTP_ENOTFOUND:
			printk("file not found");
			break;
		case TFTP_EACCESS:
			printk("access violation");
			break;
		case TFTP_ENOSPACE:
			printk("out of space");
			break;
		case TFTP_EBADOP:
			printk("illegal tftp operation");
			break;
		case TFTP_EBADID:
			printk("unknown transfer ID");
			break;
		case TFTP_EEXISTS:
			printk("file already exists");
			break;
		case TFTP_ENOUSER:
			printk("no such user");
			break;
		case TFTP_EUNDEF:
		default:
			printk("undefined");
			break;
	}
	printk(" error code\n");

	if((tftpstate->inbuflen > 4) && (packet->msg[0])) {
		printk(KERN_INFO "tftp: Error string from server is");
		if(tftpstate->inbuf[tftpstate->inbuflen])
			printk(" not terminated correctly\n");
		else
			printk(": \"%s\"\n", packet->msg);
	}
}

/*
 * Handle a data packet.
 * I haven't yet got a >32MB transfer (where the sequence number wraps around)
 * to work, but I suspect the bug is actually in netkit-tftpd.
 */
static int __init handledatapacket(struct initrd_tftp_state *tftpstate)
{
	u16 sequence;
	mm_segment_t old_fs;
	unsigned int bytes;
	struct tftp_datahdr *packet = (struct tftp_datahdr*)tftpstate->inbuf;

	sequence = ntohs(packet->seqno);
	if((tftpstate->seqno < sequence) ||
			((tftpstate->seqno - 65535) == sequence)) {
		tftpstate->seqno = sequence;
		if(!(sequence % 128)) printk(".");
#ifdef CONFIG_BLK_DEV_INITRD_TFTP_GZIP
		if(!(tftpstate->gzipped)) {
#endif
		old_fs = get_fs();
		set_fs(get_ds());
#ifndef ISHONI_ENGINE
		bytes = tftpstate->ramdisk.f_op->write(&tftpstate->ramdisk,
						packet->data,
						tftpstate->inbuflen - 4,
						&tftpstate->ramdisk.f_pos);
#else
#ifdef LINUX_BOOT_LOADER
        if ( (bytes = sys_write(tftpstate->ramfsFd, packet->data, tftpstate->inbuflen - 4)) != (tftpstate->inbuflen - 4))
            printk("Error in writing to Ramfs\n");
#endif
#endif /* ISHONI_ENGINE */
		set_fs(old_fs);
		if(bytes == -ENOSPC) {
			printk("\n" KERN_WARNING "tftp: The filesystem is too "
						"large for this RAM disk\n");
			printk(KERN_WARNING "tftp: Try booting with "
					"ramdisk_size=size_in_kb to increase "
					"the ramdisk size\n");
			return -1;
		}
#ifdef CONFIG_BLK_DEV_INITRD_TFTP_GZIP
		}
#endif
		if(tftpstate->inbuflen < 516) {
			tftpstate->finished = 1;
			printk("\n" KERN_INFO "tftp: Finished loading %d "
				"byte file\n", (((tftpstate->seqno - 1) << 9)
						+ tftpstate->errno - 4));
		}
		if((tftpstate->errno = sendack(tftpstate)))
			handleerror(tftpstate);
	}

	return 0;
}

/*
 * Handle a packet received from the server.
 */
static int __init handlepacket(struct initrd_tftp_state *tftpstate)
{
	u16 type;
	struct tftp_rqhdr *packet = (struct tftp_rqhdr*)tftpstate->inbuf;

	type = ntohs(packet->opcode);

	switch(type) {
		case TFTP_DATA:
			return(handledatapacket(tftpstate));
			break;	
		case TFTP_ERROR:
			handleerrorpacket(tftpstate);
			return -1;
			break;
		case TFTP_RRQ:
			printk(KERN_INFO "tftp: Received spurious RRQ\n");
			break;
		case TFTP_WRQ:
			printk(KERN_INFO "tftp: Received spurious WRQ\n");
			break;
		default:
			printk(KERN_INFO "tftp: Received unknown packet "
						"type %d\n", type);
			break;
	}

	return 0;
}

/*
 * Wait for a packet from the server, and handle resending the previous
 * acknowledge packet on timeout. After a given number of resends without
 * getting either a replacement packet from the server or an error to say
 * that the server has closed the port, we give up and return an error.
 */
static int __init getpacket(struct initrd_tftp_state *tftpstate)
{
	tftpstate->retries = TFTP_RESEND_MAX;

	do {
		tftpstate->timer.expires = jiffies + TFTP_PACKET_TIMEOUT;
		add_timer(&tftpstate->timer);

		while((tftpstate->errno = recvpacket(tftpstate)) == -EAGAIN);

		del_timer(&tftpstate->timer);
		if(signal_pending(current)) flush_signals(current);

		if(tftpstate->errno == -ERESTARTSYS) {
			if((tftpstate->errno = sendack(tftpstate)))
				return tftpstate->errno;
			continue;
		}

		if(tftpstate->errno < 0) return tftpstate->errno;

		if(tftpstate->lastport != tftpstate->serverport) {
			printk(KERN_INFO "tftp: Received packet from unknown "
						"port %d\n",
						htons(tftpstate->lastport));
			tftpstate->errno = TFTP_EBADID;
			if((tftpstate->errno = senderr(tftpstate)))
				return tftpstate->errno;
			continue;
		}

		return 0;

	} while(tftpstate->retries--);

	printk(KERN_INFO "tftp: Connection timed out\n");

	return -1;
}

/*
 * Wait for a response from the tftp server, be it the first data packet,
 * an error packet, or an error code such as "Connection refused"...
 */
static int __init waitforresponse(struct initrd_tftp_state *tftpstate)
{
	tftpstate->timer.expires = jiffies + TFTP_RRQ_TIMEOUT;
	add_timer(&tftpstate->timer);

	while((tftpstate->errno = recvpacket(tftpstate)) == -EAGAIN);

	del_timer(&tftpstate->timer);
	if(signal_pending(current)) flush_signals(current);

	if(tftpstate->errno == -ERESTARTSYS) {
		printk(KERN_INFO "tftp: No response from server\n");
		return -1;
	}

	if(tftpstate->errno < 0) return tftpstate->errno;

	tftpstate->serverport = tftpstate->lastport;

	return 0;
}

/*
 * Test the first data block for the presence of the gzip magic numbers.
 * If they are found and support for decompressing gzipped images is compiled
 * in, we set a flag and sanity check the header. From the header of the file.
 * If they are found and support for decompressing gzipped images is not
 * compiled in, we print an error message and return an error value.
 */
static int __init check_for_gzip(struct initrd_tftp_state *tftpstate)
{
	struct tftp_datahdr *packet = (struct tftp_datahdr*)tftpstate->inbuf;

	if((packet->data[0] == 0x1f) && (packet->data[1] == 0x8b)) {
		printk(KERN_INFO "tftp: Gzipped initrd file detected\n");
#ifdef CONFIG_BLK_DEV_INITRD_TFTP_GZIP
		if(tftpstate->inbuflen < 8) {
			printk(KERN_INFO "tftp: Gzipped file is too short\n");
			return 1;
		}
		if(packet->data[2] != 8) {
			printk(KERN_INFO "tftp: Gzipped file was compressed "
					"using an unknown method\n");
			return 1;
		}
		if(packet->data[3] & 0x02) {
			printk(KERN_INFO "tftp: Gzipped file is one part of a "
					"multi-part archive\n");
			return 1;
		}
		if(packet->data[3] & 0x20) {
			printk(KERN_INFO "tftp: Gzipped file is encrypted\n");
			return 1;
		}
		if((packet->data[3] & 0x40) ||
				(packet->data[3] & 0x80)) {
			printk(KERN_INFO "tftp: Gzipped file has reserved "
					"flag bits set\n");
			return 1;
		}
		tftpstate->gzipped = 1;
	} else tftpstate->gzipped = 0;
#else
		printk(KERN_INFO "tftp: This kernel was compiled without "
				"initrd decompression support\n");
		return 1;
	}
#endif
	return 0;
}

/*
 * The main request and receive loop.
 * We currently keep trying to contact the server for TFTP_RRQ_MAX attempts,
 * unless an error such as "Connection refused" is returned, in which case we
 * give up trying immediately. We don't handle restarting from the beginning
 * if the server drops the connection in this function; that should be done
 * in the calling function, depending on the return value.
 * The return value is 0 for success, 1 for "Couldn't contact server",
 * 2 for "Server refused access to file", 3 for "Server dropped connection",
 * 4 for "Connection timed out", 5 if a gzipped image was detected when support
 * for unzipping images wasn't included, or -errno if the sendrequest() failed.
 */
static int __init mainreceiveloop(struct initrd_tftp_state *tftpstate)
{
	tftpstate->finished = 0;
	tftpstate->seqno = 0;
	tftpstate->retries = TFTP_RRQ_MAX;
	tftpstate->serverport = htons(TFTP_SERVPORT);
	init_timer(&tftpstate->timer);
	tftpstate->timer.function = timeout;
	tftpstate->timer.data = (unsigned long)current->pid;

	do {
		if(!(tftpstate->retries--)) {
			printk(KERN_INFO "tftp: Giving up\n");
			return 1;
		}
		if((tftpstate->errno = sendrequest(tftpstate)))
			return tftpstate->errno;
	} while(waitforresponse(tftpstate));

	if(handlepacket(tftpstate))
		return 2;

	if(check_for_gzip(tftpstate))
		return 5;

#ifdef CONFIG_BLK_DEV_INITRD_TFTP_GZIP
	if(tftpstate->gzipped)
		return(start_gunzip(tftpstate));
#endif

	while(1) {
		if(!(getpacket(tftpstate))) {
			if(handlepacket(tftpstate)) return 3;
			if(tftpstate->finished) break;
		} else return 4;
	}

	return 0;
}

/*
 * Attempt to open the initial ramdisk device.
 */
static int __init openramdisk(struct initrd_tftp_state *tftpstate)
{
	if ((tftpstate->ramdisk_inode = get_empty_inode()) == NULL) {
                printk("openramdisk: error with get_empty_inode()\n");
		return -ENXIO;
	}
	memset(&tftpstate->ramdisk, 0, sizeof(tftpstate->ramdisk));
	memset(&tftpstate->ramdisk_dentry, 0, sizeof(tftpstate->ramdisk_dentry));
	tftpstate->ramdisk.f_mode = 3; /* read/write */
	tftpstate->ramdisk.f_dentry = &tftpstate->ramdisk_dentry;
	tftpstate->ramdisk_dentry.d_inode = tftpstate->ramdisk_inode;
	tftpstate->ramdisk.f_op = &def_blk_fops;
	init_special_inode(tftpstate->ramdisk_inode, 
		S_IFBLK | S_IRUSR | S_IWUSR, kdev_t_to_nr(MKDEV(RAMDISK_MAJOR, 0)));

	return blkdev_open(tftpstate->ramdisk_inode, &tftpstate->ramdisk);

}

#ifdef ISHONI_ENGINE
int tftp_start = 0;
#endif

/*
 * Attempt to load the ramdisk using the specified file on the specified
 * tftp server.
 */
#ifndef ISHONI_ENGINE
int __init initrd_tftp_load(void)
#else
#ifdef LINUX_BOOT_LOADER
int __init initrd_tftp_load(char *file)
#endif
#endif
{
	int retno;
	struct initrd_tftp_state *tftpstate;

#ifdef ISHONI_ENGINE
    int address;
    int status;
    if (tftp_start == 0) 
        return -1;
    address = getHostIPAddr();
    ipnumtostr(serverip, address);
    /* getTftpFilePathName(filename); */
    strcpy(filename, file);
#endif

	if(!filename[0]) return -1;
	if(!serverip[0]) return -1;

	tftpstate = kmalloc(sizeof(struct initrd_tftp_state), GFP_KERNEL);
	if(!tftpstate) {
		printk(KERN_ERR "tftp: couldn\'t allocate state structure\n");
		return 1;
	}

	memset(tftpstate, 0, sizeof(struct initrd_tftp_state));
#ifndef ISHONI_ENGINE
	if(!(retno = openramdisk(tftpstate))) {
#else
    if ( (status =  sys_mount(" ", "/mnt", "ramfs", 0, 0)) < 0 )
    {
        printk(" Error in mounting RAMFS status %d\n",status);
        return -1;
    }
    if (( status = createFile("/mnt/ramfs")) < 0 )
            return -1;
	if((tftpstate->ramfsFd = sys_open("/mnt/ramfs", 2, 0)) > 0) {
#endif
		if(!(retno = bindsocket(tftpstate))) {
			if(!(retno = mainreceiveloop(tftpstate)))
			/*	ROOT_DEV = MKDEV(RAMDISK_MAJOR, 0); */
			sock_release(tftpstate->sock);
		}
#ifndef ISHONI_ENGINE
	} else printk(KERN_ERR "tftp: Couldn't open ramdisk device\n");
#else
	} else printk(KERN_ERR "tftp: Couldn't open ramfile %d\n",tftpstate->ramfsFd);
    sys_close(tftpstate->ramfsFd);
#endif

	kfree(tftpstate);

	return retno;
}

/*
 * Parse the kernel command line argument to extract the server IP
 * address and/or the filename to retrieve.
 */
int __init initrd_tftp_setup(char *str)
{
	int i, n;
	char *p, *pp;

	if(!str || !*str) return 1;

	if(str[0] == 'n' && str[1] == 'o' && str[2] == 'n' && str[3] == 'e') {
		serverip[0] = 0;
		filename[0] = 0;
		return 1;
	}

	p = str;
	pp = serverip;
	i = 0;
	if(p[i] != ':') {
		while(p[i] && (p[i] != ':')) {
			pp[i] = p[i];
			if(++i == (TFTP_SERVERNAMELEN - 1)) {
				strcpy(serverip,
					CONFIG_BLK_DEV_INITRD_TFTP_SERVER);
				return 1;
			}
		}
		pp[i] = 0;
	}
	i++;
	n = i;
	pp = filename;
	if(p[i]) {
		while(p[i]) {
			pp[i - n] = p[i];
			if((++i - n) == (TFTP_FILELEN - 1)) {
				strcpy(filename,
					CONFIG_BLK_DEV_INITRD_TFTP_FILE);
				return 1;
			}
		}
		pp[i] = 0;
	}
	return 1;
}
/*
 * This macro does some GCC black magic which causes the above function
 * to be called at boot for any initrd_tftp= arguments.
 */
__setup("initrd_tftp=", initrd_tftp_setup);

/*
 * This code gets included if gzipped ramdisk support is requested.
 * I apologise for it being ugly and difficult to follow, but I wanted
 * to use the lib/inflate.c (which is taken almost directly from gzip)
 * already included with the kernel tree rather than integrate yet
 * another decompression library into the kernel.
 */
#ifdef CONFIG_BLK_DEV_INITRD_TFTP_GZIP

/* Function prototype backwards compatibility for K&R compilers: */
#define OF(args) args

/* We don't have bzero() or memzero() so we use memset() instead: */
#define memzero(s, n) memset((s), 0, (n))

/* Define Gzip's non standard data type names: */
typedef unsigned char uch;
typedef unsigned short ush;
typedef unsigned long ulg;
#define STATIC static

/* The size of the decompression window: */
#define WSIZE 0x8000

/* Pointer to the start of the input buffer (packet->data): */
static uch *inbuf __initdata;
/* Pointer to the decompression window: */
static uch *window __initdata;
/* The current number of valid bytes of input left in the input buffer: */
static unsigned int insize __initdata;
/* The number of the next byte due to be processed in the input buffer: */
static unsigned int inptr __initdata;
/* The number of bytes in the window which are ready to be written out: */
static unsigned int outcnt __initdata = 0;
/* Counts how many bytes have been decompressed (for comparison to header): */
static unsigned long bytes_out __initdata = 0;
/* Unfortunately this is needed because of fill_inbuf() and flush_window(): */
static struct initrd_tftp_state *gtftpstate __initdata;

/* Return the next byte from the input, refilling the input buffer if needed: */
#define get_byte()  (inptr < insize ? inbuf[inptr++] : fill_inbuf())

/* Diagnostic functions... #define DEBUG somewhere to activate */
#ifdef DEBUG
#define Assert(cond,msg) {if(!(cond)) error(msg);}
#define Trace(x) fprintf x
#define Tracev(x) {if (verbose) fprintf x ;}
#define Tracevv(x) {if (verbose>1) fprintf x ;}
#define Tracec(c,x) {if (verbose && (c)) fprintf x ;}
#define Tracecv(c,x) {if (verbose>1 && (c)) fprintf x ;}
#else
#define Assert(cond,msg)
#define Trace(x)
#define Tracev(x)
#define Tracevv(x)
#define Tracec(c,x)
#define Tracecv(c,x)
#endif

/* Blech... */
#include "../../lib/inflate.c"

/* Simple replacements for the standard libc malloc() and free() functions: */
static void __init *malloc(int size)
{
	return kmalloc(size, GFP_KERNEL);
}
static void __init free(void *where)
{
        kfree(where);
}

/* I'm not sure what these are supposed to do, but we don't need them: */
static void __init gzip_mark(void **ptr)
{
}
static void __init gzip_release(void **ptr)
{
}

/*
 * This is used to fill the input buffer. It is somewhat complicated here
 * by the fact that gunzip() wants to have a main loop which decompresses
 * the data, and calls functions to read new data in and write old data
 * out, which doesn't fit well into our model of also having a main receive
 * loop which reads packets from the server, writes them out, sends
 * acknowledement packets, and retries on timeout. What we do is, if the
 * gzip magic numbers are detected at the beginning of the first data block,
 * we call start_gunzip() which sets various things up and then enters the
 * main gunzip() loop. Whenever the decompression code runs out of data
 * (which it obtains via NEEDBITS() -> NEXTBYTE() -> * get_byte()),
 * fill_inbuf() gets called to refill the input buffer. Our implementation
 * of fill_inbuf() then effectively performs the same task as the end while()
 * loop of mainreceiveloop(). If a network error is encountered, fill_inbuf()
 * returns -1, which causes the decompression to abort.
 */
static int __init fill_inbuf(void)
{
	struct tftp_datahdr *packet = (struct tftp_datahdr*)gtftpstate->inbuf;
	if(gtftpstate->finished) return -1;
	if(getpacket(gtftpstate)) return -1;
	if(handlepacket(gtftpstate)) return -1;
	insize = gtftpstate->inbuflen - 4;
	if(!insize) return -1;
	inptr = 1;
	return(packet->data[0]);
}

/*
 * Write a block of decompressed data out to the ramdisk and update the
 * crc and bytes_out.
 */
static void __init flush_window(void)
{
	mm_segment_t old_fs;
	unsigned long c = crc;
	unsigned int n, bytes;
	unsigned char *in, ch;
	static int warned = 0;
	
	old_fs = get_fs();
	set_fs(get_ds());
	bytes = gtftpstate->ramdisk.f_op->write(&gtftpstate->ramdisk, window,
					outcnt, &gtftpstate->ramdisk.f_pos);
	set_fs(old_fs);

	if(bytes == -ENOSPC) {
		gtftpstate->finished = 1;
		if(!warned) {
			printk("\n" KERN_WARNING "tftp: The filesystem is too "
						"large for this RAM disk\n");
			printk(KERN_WARNING "tftp: Try booting with "
				"ramdisk_size=size_in_kb to increase "
					"the ramdisk size\n");
			warned = 1;
		}
	}

	in = window;
	for(n = 0; n < outcnt; n++) {
		ch = *in++;
		c = crc_32_tab[((int)c ^ ch) & 0xff] ^ (c >> 8);
	}
	crc = c;

	bytes_out += (unsigned long) outcnt;
	outcnt = 0;
}

/*
 * Print an error message on behalf of the decompression code, and set a
 * flag to make sure the decompression stops at the next fill_inbuf().
 */
static void __init error(char *msg)
{
	printk(KERN_ERR "%s\n", msg);
	gtftpstate->finished = 1;
}

/*
 * Set up various things needed for the decompresser then call gunzip()
 * to do the compression, then return a suitable value to indicate whether
 * we succeeded or not.
 */
static int __init start_gunzip(struct initrd_tftp_state *tftpstate)
{
	int retval;
	struct tftp_datahdr *packet = (struct tftp_datahdr*)tftpstate->inbuf;

	gtftpstate = tftpstate;

	tftpstate->ramdisk.f_pos = 0;
	inbuf = packet->data;
	insize = tftpstate->inbuflen - 4;
	inptr = 0;	

	if(!(window = kmalloc(WSIZE, GFP_KERNEL))) {
		printk(KERN_ERR "tftp: Couldn't allocate decompression "
								"window\n");
		return -ENOMEM;
	}

	makecrc();

	retval = gunzip();
	kfree(window);

	printk(KERN_INFO "tftp: Uncompressed size is %d bytes\n",
						(unsigned int) bytes_out);

	return(retval);
}
#endif
