/*
 * tmbutton.c - transmeta button device driver
 *
 * Copyright (C) 1999-2000 Transmeta Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/module.h>
#include <linux/version.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <linux/pci.h>
#include <linux/signal.h>
#include <asm/io.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <asm/segment.h>
#include <asm/types.h>
#include <linux/types.h>
#include <linux/wrapper.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>
#include <linux/vmalloc.h>
#include <linux/miscdevice.h>

#undef TMBUTTON_WORRY_ABOUT_ACPI
#define TMBUTTON_BUF_SIZE	256

static struct pci_dev *m7101 = NULL, *m1533 = NULL;
static int ali_acpi_base = 0;
static int sci_irq = 0;
static int board_rev = 0;
static unsigned char sci_irq_routing[16] = {
	13, 9, 3, 10, 4, 5, 7, 6, 1, 11, 0, 12, 0, 14, 0, 15
};

struct tmbutton_queue {
	unsigned long head;
	unsigned long tail;
	wait_queue_head_t proc_list;
	struct fasync_struct *fasync;
	unsigned char buf[TMBUTTON_BUF_SIZE];
};

static int tmbutton_users = 0;
static spinlock_t tmbutton_lock = SPIN_LOCK_UNLOCKED;
static struct tmbutton_queue *queue;	/* button data buffer */

static unsigned char get_from_queue(void)
{
	unsigned long flags;
	unsigned char result;

	spin_lock_irqsave(&tmbutton_lock, flags);
	result = queue->buf[queue->tail];
	queue->tail = (queue->tail + 1) & (TMBUTTON_BUF_SIZE - 1);
	spin_unlock_irqrestore(&tmbutton_lock, flags);
	return result;
};

static void put_in_queue(unsigned char data)
{
	unsigned long flags;
	int head;

	spin_lock_irqsave(&tmbutton_lock, flags);
	head = queue->head;
	queue->buf[head] = data;
	head = (head + 1) & (TMBUTTON_BUF_SIZE - 1);
	if (head != queue->tail) {
		queue->head = head;
		spin_unlock_irqrestore(&tmbutton_lock, flags);
		kill_fasync(&queue->fasync, SIGIO, POLL_IN);
		wake_up_interruptible(&queue->proc_list);
	} else
		spin_unlock_irqrestore(&tmbutton_lock, flags);
}

#define queue_empty()			(queue->head == queue->tail)
#define read_acpi_byte(reg)		(inb(ali_acpi_base + reg))
#define write_acpi_byte(reg,val)	(outb(val, ali_acpi_base + reg))

static void tmbutton_irq(int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned char status1;
	unsigned char status2;
	unsigned char state;


	if (board_rev < 1) {
		status1 = read_acpi_byte(0x1d);
		if (!status1)
			return;	/* not our interrupt */
		write_acpi_byte(0x1d, status1);	/* clear interrupt */
		pci_read_config_byte(m7101, 0x99, &status1);
		put_in_queue(~status1);
		return;
	} else {	/* rev b motherboard */
		status1 = read_acpi_byte(0x1b); /* for RUN_ENT1,2,3,4,6 */
		status2 = read_acpi_byte(0x1c);	/* for RUN_ENT14 */
		if (!status1 && !status2)
			return;	/* not our interrupt */
		write_acpi_byte(0x1b, status1);	/* clear interrupt */
		write_acpi_byte(0x1c, status2);	/* clear interrupt */
		pci_read_config_byte(m7101, 0x97, &status1); /* 1,2,3,4,6 */
		pci_read_config_byte(m7101, 0x98, &status2); /* 14 */
		/* create output compatible with rev A assignments */
		state = status1 & 0x1f;
		state |= ((status1 & 0x40) >> 1);
		state |= (status2 & 0x40);
		put_in_queue(~state);
		return;
	}
}

static int tmbutton_fasync(int fd, struct file *filp, int on)
{
	int retval;

	retval = fasync_helper(fd, filp, on, &queue->fasync);
	if (retval < 0)
		return retval;
	return 0;
}

static int tmbutton_open(struct inode *inode, struct file *filp)
{
	if (tmbutton_users++)
		return 0;
	if (board_rev < 1)	/* enable RUN_ENT17-23 SCI */
		write_acpi_byte(0x25, 0xfe);
	else {			/* enable RUN_ENT1,2,3,4,6,14 SCI */
		write_acpi_byte(0x24, 0x40);
		write_acpi_byte(0x23, 0x5e);
	}
	queue->head = queue->tail = 0;
	MOD_INC_USE_COUNT;
	return 0;
}

static int tmbutton_close(struct inode *inode, struct file *filp)
{
	if (--tmbutton_users)
		return 0;
	if (board_rev < 1)	/* disable RUN_ENT17-23 SCI */
		write_acpi_byte(0x25, 0);
	else { 			/* disable RUN_ENT1,2,3,4,6,14 SCI */
		write_acpi_byte(0x24, 0);
		write_acpi_byte(0x23, 0);
	}
	MOD_DEC_USE_COUNT;
	return 0;
}

static unsigned int tmbutton_poll(struct file *filp, poll_table *wait)
{
	poll_wait(filp, &queue->proc_list, wait);
	if (!queue_empty())
		return POLLIN | POLLRDNORM;
	return 0;
}

static int tmbutton_ioctl(struct inode *inode, struct file *filp,
			  unsigned int cmd, unsigned long arg)
{
	return -ENOIOCTLCMD;
}

static ssize_t tmbutton_write(struct file *filp, const char *buf, size_t count,
			     loff_t *ppos)
{
	return -EINVAL;
}

static ssize_t tmbutton_read(struct file *filp, char *buf, size_t count,
			     loff_t *pos)
{
	DECLARE_WAITQUEUE(wait, current);
	ssize_t bytes_read = 0;
	unsigned char p;

	if (queue_empty()) {	/* wait for an event */
		if (filp->f_flags & O_NONBLOCK)
			return -EWOULDBLOCK;
		add_wait_queue(&queue->proc_list, &wait);
check_again:
		set_current_state(TASK_INTERRUPTIBLE);
		if (queue_empty() && !signal_pending(current)) {
			schedule();
			goto check_again;
		}
		current->state = TASK_RUNNING;
		remove_wait_queue(&queue->proc_list, &wait);
	}
	while (bytes_read < count && !queue_empty()) {
		p = get_from_queue();
		if (copy_to_user (buf + bytes_read, (void *) &p, 1)) {
			bytes_read = -EFAULT;
			break;
		}
		bytes_read++;
	}
	if (bytes_read > 0) {
		/* filp->f_dentry->d_inode->i_atime = CURRENT_TIME; */
		return bytes_read;
	}
	if (signal_pending(current))
		return -ERESTARTSYS;

	return bytes_read;
}

static struct file_operations tmbutton_fops = {
	read: tmbutton_read,
	write: tmbutton_write,
	poll: tmbutton_poll,
	ioctl: tmbutton_ioctl,
	open: tmbutton_open,
	release: tmbutton_close,
	fasync: tmbutton_fasync,
};

static struct miscdevice tmbutton_miscdev = {
	TMBUTTON_MINOR, "tmbutton", &tmbutton_fops
};

int tmbutton_init(void)
{
	int result;
	__u32 pcidata;
	unsigned char b;

	if (!(m7101 = pci_find_device(PCI_VENDOR_ID_AL,
		PCI_DEVICE_ID_AL_M7101, m7101))) {
		printk(KERN_ERR "tmbutton: ALI M7101 not found.\n");
		return -EINVAL;
	}
	if (!(m1533 = pci_find_device(PCI_VENDOR_ID_AL,
		PCI_DEVICE_ID_AL_M1533, m1533))) {
		printk(KERN_ERR "tmbutton: ALI M1533 not found.\n");
		return -EINVAL;
	}
	printk(KERN_INFO "tmbutton: initializing button device.\n");
	/* get the base address for the ACPI config registers in IO space */
	pci_read_config_dword(m7101, 0xe0, &ali_acpi_base);
	ali_acpi_base &= 0xffff;
#ifdef TMBUTTON_WORRY_ABOUT_ACPI
	if (!(read_acpi_byte(0x4) & 1)) {
		printk(KERN_ERR "tmbutton: ACPI appears to be disabled! "
		       "Check your BIOS config.  Aborting.\n");
		return -EINVAL;
	}
#endif
	/* setup GPIO18/ACGAME6 as GPI18 to identify board rev */
	pci_read_config_dword(m1533, 0x7c, &pcidata);
	pcidata &= 0xfffffffc;
	pcidata |= 0x00000001;
	pci_write_config_dword(m1533, 0x7c, pcidata);

	pci_read_config_dword(m7101, 0xb4, &pcidata);
	if (!(pcidata & 0x00000400))
		board_rev++;
	printk(KERN_INFO "tmbutton: Rev %c motherboard detected\n",
		'A' + board_rev);

	/* we need to setup all the same registers bios should have */
	/* already setup for us because there is a chance that the */
	/* user could have a different bios that isn't just right... */

	/* clear RUN_ENT SMI */
	pci_write_config_dword(m7101, 0x4c, 0);
	pci_write_config_dword(m7101, 0x50, 0);

	if (board_rev < 1) {
		/* Disable floppy to enable EGPIO0-14 as EGPIO */
		pci_read_config_dword(m1533, 0x7c, &pcidata);
		pcidata |= 0x00000010;
		pci_write_config_dword(m1533, 0x7c, pcidata);
		/* Enable rising and falling edge irq RUN_ENT17 to RUN_ENT23 */
		pci_read_config_dword(m7101, 0x90, &pcidata);
		pcidata &= 0x0fffffff;
		pcidata |= 0x70000000;
		pci_write_config_dword(m7101, 0x90, pcidata);
		pci_read_config_dword(m7101, 0x94, &pcidata);
		pcidata &= 0xfff00000;
		pcidata |= 0x00077777;
		pci_write_config_dword(m7101, 0x94, pcidata);
		/* Set EGPIO1 to EGPIO7 as inputs RUN_ENT17 to RUN_ENT23 */
		pci_read_config_dword(m7101, 0x9c, &pcidata);
		pcidata &= 0x010000000;
		pci_write_config_dword(m7101, 0x9c, pcidata);
	} else {
		/* Enable rising and falling edge irq RUN_ENT1 */
		pci_read_config_dword(m7101, 0x88, &pcidata);
		pcidata &= 0x0fffffff;
		pcidata |= 0x70000000;
		pci_write_config_dword(m7101, 0x88, pcidata);
		/* Enable rising and falling edge irq RUN_ENT2,3,4,6 */
		pci_read_config_dword(m7101, 0x8c, &pcidata);
		pcidata &= 0xfff0f000;
		pcidata |= 0x00070777;
		pci_write_config_dword(m7101, 0x8c, pcidata);
		/* Enable rising and falling edge irq RUN_ENT14 */
		pci_read_config_dword(m7101, 0x90, &pcidata);
		pcidata &= 0xfff0ffff;
		pcidata |= 0x00070000;
		pci_write_config_dword(m7101, 0x90, pcidata);
		/* Set GPIO1,2,3,4,6,14 as inputs RUN_ENT1,2,3,4,6,14 */
		pci_read_config_dword(m7101, 0x98, &pcidata);
		pcidata &= 0x4005e0000;
		pci_write_config_dword(m7101, 0x98, pcidata);
	}
	/* Get System Control Interrupt IRQ routing */
	pci_read_config_byte(m1533, 0x76, &b);
	if ((b & 0x8f) == 0x80)	/* if SCI Routing to IRQ 13 disable set */
		b = 14;		/* then setup a fallthrough to bail out */
	if (!(b = sci_irq_routing[b & 0x0f])) {
		printk(KERN_ERR "tmbutton: SCI IRQ is disabled. Aborting.\n");
		return -EINVAL;
	}
	sci_irq = b;
	printk(KERN_INFO "tmbutton: requesting SCI interrupt, IRQ %d\n", b);
	result = request_irq(sci_irq, tmbutton_irq, SA_SHIRQ | SA_INTERRUPT,
		"tmbutton", &queue);
	if (result == -EINVAL)
		printk(KERN_ERR "tmbutton: bad irq number or handler\n");
	if (result == -EBUSY)
		printk(KERN_ERR "tmbutton: IRQ %d in use and not shared\n",
			sci_irq);
	if (result < 0)
		return result;
	queue = (struct tmbutton_queue *) kmalloc(sizeof(*queue), GFP_KERNEL);
	if (!queue) {
		free_irq(sci_irq, &queue);
		printk(KERN_ERR "tmbutton: kmalloc failed.  Aborting.\n");
		return -ENOMEM;
	}
	memset(queue, 0, sizeof(*queue));
	/* set SCI IRQ to LEVEL triggered if bios failed to do so */
	if (sci_irq < 8)
		b = inb(0x4d0) & (1<<sci_irq);
	else
		b = inb(0x4d1) & (1<<(sci_irq - 8));
	if (!b) {
		printk(KERN_ERR "tmbutton: fixing IRQ %d to LEVEL trigger. "
		       " Your BIOS is buggy.\n", sci_irq);
		if (sci_irq < 8) {
			b = inb(0x4d0) | (1<<sci_irq);
			outb(b, 0x4d0);
		} else {
			b = inb(0x4d1) | (1<<(sci_irq - 8));
			outb(b, 0x4d1);
		}
	}
	/* don't forget to init the waitqueue or else we oops on read */
	init_waitqueue_head(&queue->proc_list);
	/* register under major 10, minor specified in tmbutton_miscdev */
	misc_register(&tmbutton_miscdev);
	printk(KERN_INFO "tmbutton: device registered\n");
	return 0;
}

#ifdef MODULE

MODULE_AUTHOR("Nathan Laredo");
MODULE_DESCRIPTION("Transmeta button device driver");

int init_module(void)
{
	printk(KERN_INFO "tmbutton: installing module\n");
	return tmbutton_init();
}

void cleanup_module(void)
{
	misc_deregister(&tmbutton_miscdev);
	free_irq(sci_irq, &queue);
	kfree(queue);
	printk(KERN_INFO "tmbutton: module uninstalled\n");
}

#endif
