
/*
 * linux/drivers/char/vrc4173.c
 *
 * Driver for NEC Vrc4173 core.
 *
 * Copyright (C) 2000 Michael R. McDonald
 * Copyright (C) 2001 Montavista Software Inc.
 * Author: Yoichi Yuasa
 *		yyuasa@mvista.com or source@mvista.com
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 */

#include <linux/module.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/delay.h>

#include <asm/io.h>
#include <asm/vrc4173.h>

static const char *version = "NEC Vrc4173 driver ver 2.0\n";

/*
 * PCI device identifiers for "new style" Linux PCI Device Drivers
 */
static struct pci_device_id vrc4173_pci_tbl[] __devinitdata = {
	{PCI_VENDOR_ID_NEC,PCI_DEVICE_ID_NEC_VRC4173_BCU,PCI_ANY_ID,PCI_ANY_ID,0,0,0},
	{0,}
};

MODULE_DEVICE_TABLE(pci, vrc4173_pci_tbl);

struct pci_dev *vrc4173_dev = NULL;
unsigned long vrc4173_io_port_base = 0;

/*===============================================================*/

spinlock_t vrc4173_icu_lock = SPIN_LOCK_UNLOCKED;

void enable_vrc4173_irq(unsigned short mask)
{
        unsigned long flags;
        u16 val;

        spin_lock_irqsave(&vrc4173_icu_lock, flags);
        val = vrc4173_inw(VRC4173_MSYSINT1REG);
        val |= mask;
        vrc4173_outw(val, VRC4173_MSYSINT1REG);
        spin_unlock_irqrestore(&vrc4173_icu_lock, flags);
}

void disable_vrc4173_irq(unsigned short mask)
{
        unsigned long flags;
        u16 val;

        spin_lock_irqsave(&vrc4173_icu_lock, flags);
        val = vrc4173_inw(VRC4173_MSYSINT1REG);
        val &= ~mask;
        vrc4173_outw(val, VRC4173_MSYSINT1REG);
        spin_unlock_irqrestore(&vrc4173_icu_lock, flags);
}

static void vrc4173_icu_init(void)
{
	vrc4173_outw(0, VRC4173_MSYSINT1REG);
	vrc4173_outw(0, VRC4173_MPIUINTREG);
	vrc4173_outw(0, VRC4173_MAIUINTREG);
	vrc4173_outw(0, VRC4173_MKIUINTREG);
	vrc4173_outw(0, VRC4173_MGIULINTREG);
	vrc4173_outw(0, VRC4173_MGIUHINTREG);

	vrc4173_outw(0, VRC4173_GIUINTENL);
	vrc4173_outw(0, VRC4173_GIUINTENH);

	vrc4173_outw(0xffff, VRC4173_GIUINTSTATL);
	vrc4173_outw(0xffff, VRC4173_GIUINTSTATH);
}

/*===============================================================*/

spinlock_t vrc4173_cmu_lock = SPIN_LOCK_UNLOCKED;

void vrc4173_clock_supply(unsigned short mask)
{
	unsigned long flags;
	unsigned short val;

	write_lock_irqsave(&vrc4173_cmu_lock, flags);
	val = vrc4173_inw(VRC4173_CMUCLKMSK);
        val |= mask;
        vrc4173_outw(val, VRC4173_CMUCLKMSK);
        write_unlock_irqrestore(&vrc4173_cmu_lock, flags);
}

void vrc4173_clock_mask(unsigned short mask)
{
	unsigned long flags;
	unsigned short val;

	write_lock_irqsave(&vrc4173_cmu_lock, flags);
	val = vrc4173_inw(VRC4173_CMUCLKMSK);
	val &= ~mask;
	vrc4173_outw(val, VRC4173_CMUCLKMSK);
	write_unlock_irqrestore(&vrc4173_cmu_lock, flags);
}

static void vrc4173_cmu_init(void)
{
	unsigned short val;

	/* Enable 48MHz */
	vrc4173_outw(VRC4173_CMUCLKMSK_MSK48MOSC, VRC4173_CMUCLKMSK);
	udelay(100);
	val = vrc4173_inw(VRC4173_CMUCLKMSK);
	val |= VRC4173_CMUCLKMSK_MSK48MPIN;
	vrc4173_outw(val, VRC4173_CMUCLKMSK);

	/* Reset CMU  */
	vrc4173_outw(0x000f, VRC4173_CMUSRST);
	udelay(1000);
	vrc4173_outw(0, VRC4173_CMUSRST);
}

/*===============================================================*/

static int __init vrc4173_pci_probe(struct pci_dev *pdev,
                                    const struct pci_device_id *ent)
{
	long ioaddr;
	int err = 0;

	printk(KERN_INFO "%s", version);
	printk(KERN_INFO "    found device %#08x.%#08x\n", ent->vendor, ent->device);

	if ((err = pci_enable_device(pdev)) < 0) {
		printk(KERN_ERR "NEC Vrc4173: failed to enable device -- err=%d\n",
		       err);
		return err;
	}

	pci_set_master(pdev);

	ioaddr = pci_resource_start (pdev, 0);
	if (!ioaddr) {
		printk (KERN_ERR "no PCI IO resources, aborting\n");
		return -ENODEV;
	}

	if (!ioaddr || ((pci_resource_flags(pdev, 0) & IORESOURCE_IO) == 0)) {
		printk (KERN_ERR "NEC Vrc4173: no I/O resource at PCI BAR #0\n");
		return -ENODEV;
	}

	printk(KERN_INFO "    ioaddr=%#08lx  resource_flags=%#08lx\n",
	       ioaddr, pci_resource_flags (pdev, 0));


	if (request_region(ioaddr, VRC4173_IO_EXTENT, "NEC Vrc4173") == NULL) {
		printk(KERN_ERR "NEC Vrc4173: I/O resource 0x%x @ 0x%lx busy\n",
		       VRC4173_IO_EXTENT, ioaddr);
		return -EBUSY;
	}
	vrc4173_io_port_base = ioaddr;

	vrc4173_icu_init();
	vrc4173_cmu_init();

	/* setup how the GIU lines are used */
	vrc4173_outw(VRC4173_SELECTREG_MASK, VRC4173_SELECTREG);

	if (vrc4173_dev == NULL) {
		vrc4173_dev = pdev;
	}

#ifdef CONFIG_VRC4173_PSKEYBOARD
	(void)vrc4173_ps2u_init();
#endif

#ifdef CONFIG_VRC4173_PSKEYBOARD
	(void)vrc4173_tpanel_init();
#endif

	return 0;
}

static struct pci_driver vrc4173_driver = {
	name:		"NEC Vrc4173",
	probe:		vrc4173_pci_probe,
	remove:		NULL,
	id_table:	vrc4173_pci_tbl,
};

static int __init vrc4173_pci_init(void)
{
	int err;

	if ((err = pci_module_init(&vrc4173_driver)) < 0)
		return err;
	return 0;
}

static void __exit vrc4173_pci_cleanup(void)
{
	pci_unregister_driver (&vrc4173_driver);
}

module_init(vrc4173_pci_init);
module_exit(vrc4173_pci_cleanup);
