/*
 * i2c-pfs168.c
 *
 * Copyright (C) 2001 MontaVista Software, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *
 * Simple i2c access for the Radisys PFS-168 StrongArm platform.
 *
 */

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ioport.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/stddef.h>
#include <linux/parport.h>
#include <linux/rtc.h>

#include <linux/i2c.h>
#include <linux/i2c-algo-bit.h>

#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/system.h>

#include "m41t11.h"


#undef	DEBUG_M41T11
#if	defined(DEBUG_M41T11)
#	define	DPRINTK(fmt, args...)	printk(KERN_DEBUG "%s: " fmt, __FUNCTION__ , ## args)
#else
#	define	DPRINTK(fmt, args...)
#endif                                                                                                                                        


#define L3_DataPin	GPIO_GPIO(15)
#define L3_ClockPin	GPIO_GPIO(18)
#define L3_ModePin	GPIO_GPIO(17)


static int iic_acquired;
static int iic_data_out;
static int iic_clock_out;


extern unsigned long
mktime(unsigned int year, unsigned int mon, unsigned int day,
       unsigned int hour, unsigned int min, unsigned int sec);
extern int (*set_rtc)(void);

static struct i2c_client *rtc_client;

static inline int rtc_command(int cmd, void *data)
{
	int ret = -EIO;

	if (rtc_client) {
		ret = rtc_client->driver->command(rtc_client, cmd, data);
	}

	return ret;
}

/*
 * Read the current RTC time and date, and update xtime.
 */
static void get_rtc_time(void)
{
	unsigned char ctrl;
	struct rtc_time rtctm;

#if	0
	/*
	 * Ensure that the RTC is running.
	 */
	rtc_command(RTC_GETCTRL, &ctrl);
	if (ctrl & 0xc0) {
		unsigned char new_ctrl;

		new_ctrl = ctrl & ~0xc0;

		printk("RTC: resetting control %02X -> %02X\n",
		       ctrl, new_ctrl);

		rtc_command(RTC_SETCTRL, &new_ctrl);
	}
#endif

	if (rtc_command(RTC_GETDATETIME, &rtctm)) {
		return;
	}

	if (rtctm.tm_year < 70)
		rtctm.tm_year += 100;

	xtime.tv_usec = 0;
	xtime.tv_sec  = mktime(1900 + rtctm.tm_year, rtctm.tm_mon, rtctm.tm_mday,
			       rtctm.tm_hour, rtctm.tm_min, rtctm.tm_sec);
}

/*
 * Set the RTC time only.  Note that
 * we do not touch the date.
 */
static int set_rtc_time(void)
{
	struct rtc_time new_rtctm, old_rtctm;
	unsigned long nowtime = xtime.tv_sec;

	if (rtc_command(RTC_GETDATETIME, &old_rtctm))
		return 0;

	new_rtctm.tm_sec  = nowtime % 60;	nowtime /= 60;
	new_rtctm.tm_min  = nowtime % 60;	nowtime /= 60;
	new_rtctm.tm_hour = nowtime % 24;

	/*
	 * avoid writing when we're going to change the day
	 * of the month.  We will retry in the next minute.
	 * This basically means that if the RTC must not drift
	 * by more than 1 minute in 11 minutes.
	 *
	 * [ rtc: 1/1/2000 23:58:00, real 2/1/2000 00:01:00,
	 *   rtc gets set to 1/1/2000 00:01:00 ]
	 */
	if ((old_rtctm.tm_hour == 23 && old_rtctm.tm_min  == 59) ||
	    (new_rtctm.tm_hour == 23 && new_rtctm.tm_min  == 59))
		return 1;

	return rtc_command(RTC_SETTIME, &new_rtctm);
}


/*
 * Grab control of the IIC/L3 shared pins
 */
static inline void iic_acquirepins(void)
{
	GPDR |= L3_ClockPin;
	GPSR = L3_ClockPin;
	GPDR &= ~L3_DataPin;
	GPCR = L3_DataPin;
	iic_acquired = 1;
	iic_clock_out = 1;
	iic_data_out = 0;
}

/*
 * Release control of the IIC/L3 shared pins
 */
static inline void iic_releasepins(void)
{
	GPDR &= ~(L3_ClockPin | L3_DataPin);
	GPCR = (L3_ClockPin | L3_DataPin);
	iic_acquired = 0;
}

/*
 * Initialize the interface
 */
static void iic_init(void)
{
	GAFR &= ~(L3_DataPin | L3_ClockPin);
	GPSR = L3_ModePin;
	GPDR |= L3_ModePin;
	iic_releasepins();
}


/* --------- */

static void bit_pfs168_setscl(void *data, int state)
{
	if (!iic_acquired)
		iic_acquirepins();

	if (!iic_clock_out) {
		GPDR |= L3_ClockPin;
		iic_clock_out = 1;
	}

	if (state) {
		GPSR = L3_ClockPin;
	} else {
		GPCR = L3_ClockPin;
	}
	
}

static void bit_pfs168_setsda(void *data, int state)
{
	if (!iic_acquired)
		iic_acquirepins();

	if (state) {
		GPDR &= ~(L3_DataPin);
	} else {
		GPCR = L3_DataPin;
		GPDR |= L3_DataPin;
	}
	
} 

static int bit_pfs168_getscl(void *data)
{
	if (!iic_acquired)
		iic_acquirepins();

	if (iic_clock_out) {
		GPDR &= ~(L3_ClockPin);
		iic_clock_out = 0;
	}

	return (GPLR & L3_ClockPin) ? 1 : 0;
}

static int bit_pfs168_getsda(void *data)
{
	if (!iic_acquired)
		iic_acquirepins();

	if (iic_data_out) {
		GPDR &= ~(L3_DataPin);
		iic_data_out = 0;
	}

	return (GPLR & L3_DataPin) ? 1 : 0;
}

/* ---------- */

static int bit_pfs168_reg(struct i2c_client *client)
{
	if (client->id == I2C_DRIVERID_M41T11 &&
	    client->addr == 0x68) {
		rtc_client = client;
		get_rtc_time();
		set_rtc = set_rtc_time;
	}

	return 0;
}

static int bit_pfs168_unreg(struct i2c_client *client)
{
	if (client == rtc_client) {
		set_rtc = NULL;
		rtc_client = NULL;
	}

	return 0;
}

static void bit_pfs168_inc_use(struct i2c_adapter *adap)
{
#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif
	iic_acquirepins();
}

static void bit_pfs168_dec_use(struct i2c_adapter *adap)
{
	iic_releasepins();
#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif
}

static struct i2c_algo_bit_data bit_pfs168_data = {
	NULL,
	bit_pfs168_setsda,
	bit_pfs168_setscl,
	bit_pfs168_getsda,
	NULL,
	5, 5, 100,		/*	waits, timeout */
};


static struct i2c_adapter bit_pfs168_ops = {
	"PFS-168",
	I2C_HW_B_IOC,
	NULL,
	&bit_pfs168_data,
	bit_pfs168_inc_use,
	bit_pfs168_dec_use,
	bit_pfs168_reg,
	bit_pfs168_unreg,
};

int __init i2c_pfs168_init(void)
{
	printk("i2c-pfs168.o: i2c PFS-168 module\n");

	/* reset hardware to sane state */
	iic_init();

	if (i2c_bit_add_bus(&bit_pfs168_ops) < 0) {
		printk("i2c-pfs168: Unable to register with I2C\n");
		return -ENODEV;
	}

	return 0;
}

void __exit i2c_pfs168_exit(void)
{
	i2c_bit_del_bus(&bit_pfs168_ops);
}


unsigned char adv7171_settings[] = {
  0x00, /* subaddr */
  0x00, /* Mode Register 0 */
  0x80, /* Mode Register 1 */
  0x08, /* Mode Register 2 */
  0x00, /* Mode Register 3 */
  0x18, /* Mode Register 4 */
  0x00, /* Reserved */
  0x00, /* Reserved */
  0x4c, /* Timing Register 0 */
  0x00, /* Timing Register 1 */
  0x16, /* Subcarrier Freq 0 */
  0x7c, /* Subcarrier Freq 1 */
  0xf0, /* Subcarrier Freq 2 */
  0x21, /* Subcarrier Freq 3 */
  0x00, /* Subcarrier phase */
  0x00, /* Extended Captioning 0 */
  0x00, /* Extended Captioning 1 */
  0x00, /* Closed Captioning 0 */
  0x00, /* Closed Captioning 1 */
  0x00, /* Pedestal Register 0 */
  0x00, /* Pedestal Register 1 */
  0x00, /* Pedestal Register 2 */
  0x00, /* Pedestal Register 3 */
  0x00, /* CGMS_WSS 0 */
  0x00, /* CGMS_WSS 1 */
  0x00, /* CGMS_WSS 2 */
  0x00  /* Teletext request pos */
};

int pfs168_i2c_adv7171_init(void)
{
  struct i2c_adapter *adap = &bit_pfs168_ops;
  struct i2c_msg msg[1];
  int ret;

  printk("i2c_pfs168: adv7171 init\n");

  msg[0].addr = 0x54 / 2;
  msg[0].flags = 0;
  msg[0].len = 26+1;
  msg[0].buf = adv7171_settings;

  printk("starting hack\n");
  ret = adap->algo->master_xfer(adap, msg, 1);
  printk("hack ret %d\n", ret);

  printk("i2c_pfs168: adv7171 init done\n");
  return 0;
}

MODULE_AUTHOR("George G. Davis <davis_g@mvista.com>");
MODULE_DESCRIPTION("I2C-Bus adapter routines for PFS-168");

module_init(i2c_pfs168_init);
module_exit(i2c_pfs168_exit);
