/*
 * cpx1024.c
 *
 * Driver to control access to the hardware monitor on the CPX1024 Quad
 * Xeon processor machine built by Diversified Technology Inc.
 * 
 * For information on programming this driver, refer to
 * Documentation/cpx1024.txt in the Linux kernel source code.
 *
 * Author: Johnnie Peters
 *         jpeters@mvista.com
 *
 * Copyright 2000 MontaVista Softare Inc.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR   IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT,  INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/malloc.h>
#include <linux/ioctl.h>
#include <linux/wait.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <linux/cpx1024mon.h>

#include "cpx1024.h"

static ssize_t cpx1024mon_read(struct file *, char *, size_t, loff_t *);
static int     cpx1024mon_ioctl(struct inode *, struct file *, unsigned int, unsigned long);
static int     cpx1024mon_open(struct inode *, struct file *);
static int     cpx1024mon_release(struct inode *, struct file *);

void           cpx1024mon_regread(unsigned char, int *, int *, int);
void           cpx1024mon_regread2(unsigned char, int, int *, int *, int);
void	       cpx1024mon_regwrite(unsigned char, int, int, int, int);

void	       cpx1024mon_poll(unsigned long);
void           cpx1024mon_send_all_event(void);

void           cpx1024mon_sendevent(int, int, int);

void           cpx1025mon_setuid(char *);

static struct file_operations cpx1024mon_fops = {
	read:		cpx1024mon_read,
	ioctl:		cpx1024mon_ioctl,
	open:		cpx1024mon_open,
	release:	cpx1024mon_release,
};

/* Structures to control polling timeout */
struct timer_list cpx1024mon_timer;
int cpx1024mon_timeout_interval = 1;

/* User list structures */
struct cpx1024mon_user cpx1024monUserTab[CPX1024_MAX_USERS];
struct cpx1024mon_user *cpx1024monUserFreeList;
struct cpx1024mon_user *cpx1024monUserList;
int cpx1024mon_user_cnt = 0;

/* List of free messages buffers */
struct cpx1024mon_message *cpx1024monMessFreeList;

/*
 * Linux driver initialization entry point
 */
int __init
cpx1024mon_init(void)
{
	unsigned char status;
	int mon_rev;
	int mon_patch;
	int bios_rev;
	int cpu_mask;
	int major;
	int sys_status;
	int loop;
	struct cpx1024mon_message *tmpMessPtr;

	/* Check monitor for strange behavior */
	status = inb(DTIMON_CMD);
	if (!(status & DTIMON_STATUS_CBE)) {
		printk("Cannot initialize driver, command buffer not empty\n");
		return -ENODEV;
	}

	/* Reset the monitor */
	cpx1024mon_regwrite(DTIMON_CMD_SOFTRESET, 0, 0, 0, 0);

	/* Check the status for interrupt bit */
	status = inb(DTIMON_CMD);

	if (status & DTIMON_STATUS_INTP) {
		/* Clear the interrupt */
		cpx1024mon_regread(DTIMON_CMD_GET_STATUS, &sys_status, NULL, 1);
	}

	cpx1024mon_regread(DTIMON_CMD_GET_BIOSREV, &bios_rev, NULL, 1);
	cpx1024mon_regread(DTIMON_CMD_GET_MONREV, &mon_rev, &mon_patch, 2);
	cpx1024mon_regread(DTIMON_CMD_GET_CPUMASK, &cpu_mask, NULL, 1);

	if (mon_patch == 1) {
		printk("DTI CPX1024 Hardware Monitor Rev %d.%d ",
						(mon_rev & 0xf0) >> 4, mon_rev & 0x0f);
	} else {
		printk("DTI CPX1024 Hardware Monitor Rev %d.%d Patch %d ",
						(mon_rev & 0xf0) >> 4, mon_rev & 0x0f, mon_patch);
	}
	printk("\n");

	printk("    On Bios Rev %d.%d\n", (bios_rev & 0xf0) >> 4, bios_rev & 0xf);
	printk("    CPUs installed: ");
	for (loop = 0; loop < 4; loop++) {
		if (cpu_mask & (1 << loop)) {
			printk("%d ", loop);
		}
	}
	printk("\n");

	if ((major = register_chrdev(0,"CPX1024 Hardware Monitor",
			&cpx1024mon_fops)) < 0){
		printk("Failed to register hardware monitor driver (%d)\n", major);
		return -ENODEV;
	}

	/* Intialize the user free list */
	for (loop = 0; loop < CPX1024_MAX_USERS; loop ++) {
		cpx1024monUserTab[loop].next =
			(struct cpx1024mon_user *)&cpx1024monUserTab[loop + 1].pid;
		cpx1024monUserTab[loop].pid = 0;
		init_waitqueue_head(&cpx1024monUserTab[loop].wq);
	}
	cpx1024monUserTab[CPX1024_MAX_USERS - 1].next = (struct cpx1024mon_user *)NULL;

	cpx1024monUserFreeList = cpx1024monUserTab;
	cpx1024monUserList = (struct cpx1024mon_user *)NULL;

	/* Initialize the first chunk of memory for the return free list */
	cpx1024monMessFreeList = kmalloc(CPX1024_MESSAGE_CHUNK_SIZE *
					 sizeof(struct cpx1024mon_message), GFP_ATOMIC);

	tmpMessPtr = cpx1024monMessFreeList;
	for (loop = 0; loop < CPX1024_MESSAGE_CHUNK_SIZE - 1; loop++) {
		tmpMessPtr->code = 0;
		tmpMessPtr->next = tmpMessPtr + 1;
		tmpMessPtr++;
	}
	tmpMessPtr->code = 0;
	tmpMessPtr->next = (struct cpx1024mon_message *)NULL;

	/* Set up the driver to poll the monitor hardware.  When interrupts
	 * start working they should be used instead.
	 */
	init_timer(&cpx1024mon_timer);
	cpx1024mon_timer.function = cpx1024mon_poll;
	cpx1024mon_timer.data = (unsigned long)NULL;
	cpx1024mon_timer.expires = jiffies + cpx1024mon_timeout_interval * HZ;
	add_timer(&cpx1024mon_timer);
	
	return 0;
}

/*
 * Linux read entry point is a noop.
 */
static ssize_t
cpx1024mon_read(struct file *file, char *buffer, size_t count, loff_t *ppos)
{
	return (ssize_t)0;
}

/*
 * Linux ioctl driver entry point.
 */
static int
cpx1024mon_ioctl(struct inode *inode,
				struct file *file, unsigned int cmd, unsigned long arg)
{
	struct cpx1024mon_msg_ioctl msg_buf;
	struct cpx1024mon_user *tmpUserPtr;
	struct cpx1024mon_user *delUserPtr;
	struct cpx1024mon_message *tmpMsgPtr;
	struct cpx1024mon_message *freeMsgPtr;
	int msg_num = 0;
	int temp;
	int overheat;
	int tt_mode;
	int volt;
	int amp;
	int fan;
	int gpo;
	int gpi;
	int tibm0;
	int tibm1;
	int cpumask;
	int echo_mode;
	int transfer;
	int verbose;
	unsigned long flags;
	struct cpx1024mon_temps_ioctl temp_ioctl;
	struct cpx1024mon_bound_temps_ioctl bt_ioctl;
	struct cpx1024mon_flash_ioctl flash_ioctl;
	struct cpx1024mon_force_gpo_ioctl gpo_ioctl;
	struct cpx1024mon_volt_ioctl volt_ioctl;
	struct cpx1024mon_vtoler_ioctl vtoler_ioctl;
	struct cpx1024mon_amp_ioctl amp_ioctl;
	struct cpx1024mon_fan_ioctl fan_ioctl;
	struct cpx1024mon_gpo_pol_ioctl pol_ioctl;
	struct cpx1024mon_gpi_pol_ioctl ipol_ioctl;
	struct cpx1024mon_info minfo_ioctl;
	struct cpx1024mon_passwd_ioctl passwd_ioctl;
	struct cpx1024mon_uid_ioctl sysuid_ioctl;

	switch (cmd) {
	case CPX1024MON_GET_TEMPS:
		save_flags(flags);
		cli();

		for (temp = 0; temp < CPX1024_TEMP_MAX; temp++) {
			cpx1024mon_regread2(DTIMON_CMD_GET_TEMPVAL, temp,
					    &temp_ioctl.temps[temp][0],
					    &temp_ioctl.temps[temp][1], 2);
		}

		restore_flags(flags);

		/* Copy the information back to user space */
		if(copy_to_user((void *)arg,(const void *)&temp_ioctl,
				sizeof(struct cpx1024mon_temps_ioctl )) > 0) {
			printk("cpx1024mon_ioctl: Bad copy out on temps ioctl\n");
			return -EFAULT;
		}

		break;

	case CPX1024MON_GET_BOUND_TEMPS:
		save_flags(flags);
		cli();

		for (temp = 0; temp < CPX1024_TEMP_MAX; temp++) {
			cpx1024mon_regread2(DTIMON_CMD_GET_LOWTEMP, temp,
					    &bt_ioctl.bound_temps[temp].low[0],
					    &bt_ioctl.bound_temps[temp].low[1], 2);
			cpx1024mon_regread2(DTIMON_CMD_GET_HIGHTEMP, temp,
					    &bt_ioctl.bound_temps[temp].high[0],
					    &bt_ioctl.bound_temps[temp].high[1], 2);
		}

		restore_flags(flags);

		/* Copy the information back to user space */
		if(copy_to_user((void *)arg,(const void *)&bt_ioctl,
				sizeof(struct cpx1024mon_bound_temps_ioctl )) > 0) {
			printk("cpx1024mon_ioctl: Bad copy out on bound temps ioctl\n");
			return -EFAULT;
		}

		break;

	case CPX1024MON_SET_BOUND_TEMPS:
		/* Get information from user space */
		if (copy_from_user((void *)&bt_ioctl, (const void *)arg,
				   sizeof(struct cpx1024mon_bound_temps_ioctl)) > 0) {
			printk("cpx1024mon_ioctl: Bad copy in on ioctl bound set ioctl\n");
			return -EFAULT;
		}

		save_flags(flags);
		cli();

		for (temp = 0; temp < CPX1024_TEMP_MAX; temp++) {
			if (bt_ioctl.bound_temps[temp].flag & CPX1024MON_SET_BT_UPPER) {
				cpx1024mon_regwrite(DTIMON_CMD_SET_HIGHTEMP, temp,
						    bt_ioctl.bound_temps[temp].high[0],
						    bt_ioctl.bound_temps[temp].high[1], 3);
			}
			if (bt_ioctl.bound_temps[temp].flag & CPX1024MON_SET_BT_LOWER) {
				cpx1024mon_regwrite(DTIMON_CMD_SET_LOWTEMP, temp,
						    bt_ioctl.bound_temps[temp].low[0],
						    bt_ioctl.bound_temps[temp].low[1], 3);
			}
		}

		restore_flags(flags);

		break;

	case CPX1024MON_SET_LED:
		if (copy_from_user((void *)&flash_ioctl, (const void *)arg,
				   sizeof(struct cpx1024mon_flash_ioctl)) > 0) {
			printk("cpx1024mon_ioctl: Bad copy in on flash ioctl\n");
			return -EFAULT;
		}

		save_flags(flags);
		cli();

		if (flash_ioctl.flags & CPX1024MON_LED_ACTIVE) {
			switch (flash_ioctl.active[1]) {
			case DTIMON_LED_FLASH:
				cpx1024mon_regwrite(DTIMON_CMD_SET_LEDFLASH, DTIMON_LED_ACTIVE,
					    flash_ioctl.active[0], 0, 2);
			case DTIMON_LED_OFF:
			case DTIMON_LED_ON:
				cpx1024mon_regwrite(DTIMON_CMD_ACTIVE_LED,
						    flash_ioctl.active[1], 0, 0, 1);
				break;
			default:
				restore_flags(flags);
				printk("cpx1024mon_ioctl: invalid LED control option\n");
				return -EINVAL;
			}
		}

		if (flash_ioctl.flags & CPX1024MON_LED_OOS) {
			switch (flash_ioctl.out_of_service[1]) {
			case DTIMON_LED_FLASH:
				cpx1024mon_regwrite(DTIMON_CMD_SET_LEDFLASH, DTIMON_LED_OOS,
						    flash_ioctl.out_of_service[0], 0, 2);
			case DTIMON_LED_OFF:
			case DTIMON_LED_ON:
				cpx1024mon_regwrite(DTIMON_CMD_OOS_LED,
						    flash_ioctl.out_of_service[1], 0, 0, 1);
				break;
			default:
				restore_flags(flags);
				printk("cpx1024mon_ioctl: invalid LED control option\n");
				return -EINVAL;
			}
		}

		restore_flags(flags);

		break;

	case CPX1024MON_GET_OVERHEAT:
		save_flags(flags);
		cli();

		cpx1024mon_regread(DTIMON_CMD_CPU_OVERHEAT, &overheat, NULL, 1);

		restore_flags(flags);

		/* Copy the information back to user space */
		if(copy_to_user((void *)arg,(const void *)&overheat, sizeof(int)) > 0) {
			printk("cpx1024mon_ioctl: Bad copy out on overheat ioctl\n");
			return -EFAULT;
		}

		break;

	case CPX1024MON_SOFT_RESET:
		save_flags(flags);
		cli();

		cpx1024mon_regwrite(DTIMON_CMD_SOFTRESET, 0, 0, 0, 0);

		restore_flags(flags);

		break;

	case CPX1024MON_SET_TT_MODE:
		if (copy_from_user((void *)&tt_mode, (const void *)arg, sizeof(int)) > 0) {
			printk("cpx1024mon_ioctl: Bad copy in on term trans mode ioctl\n");
			return -EFAULT;
		}

		save_flags(flags);
		cli();

		if (tt_mode) {
			cpx1024mon_regwrite(DTIMON_CMD_ENABLE_TTM, 0, 0, 0, 0);
		} else {
			cpx1024mon_regwrite(DTIMON_CMD_DISABLE_TTM, 0, 0, 0, 0);
		}

		restore_flags(flags);

		break;

	case CPX1024MON_FORCE_GPO:
		if (copy_from_user((void *)&gpo_ioctl, (const void *)arg,
				   sizeof(struct cpx1024mon_force_gpo_ioctl)) > 0) {
			printk("cpx1024mon_ioctl: Bad copy in on force gpo ioctl\n");
			return -EFAULT;
		}

		if ((gpo_ioctl.gpo < 0) || (gpo_ioctl.gpo > CPX1024_MAX_GPO)) {
			printk("cpx1024mon_ioctl: invalid GPO number\n");
			return -EINVAL;
		}

		save_flags(flags);
		cli();

printk("CPX1024MON_FORCE_GPO: for gpo %d state indicated is %d\n", gpo_ioctl.gpo, gpo_ioctl.state);
		switch (gpo_ioctl.state) {
		case CPX1024MON_GPO_LOW:
			cpx1024mon_regwrite(DTIMON_CMD_FORCE_GPO, gpo_ioctl.gpo, DTIMON_FGPO_LOW, 0, 2);
			break;
		case CPX1024MON_GPO_HIGH:
			cpx1024mon_regwrite(DTIMON_CMD_FORCE_GPO, gpo_ioctl.gpo, DTIMON_FGPO_HIGH, 0, 2);
			break;
		default:
			restore_flags(flags);
			printk("cpx1024mon_ioctl: invalid GPO state\n");
			return -EINVAL;
		}

		restore_flags(flags);

		break;

	case CPX1024MON_SET_GPO_POL:
		if (copy_from_user((void *)&pol_ioctl, (const void *)arg,
				   sizeof(struct cpx1024mon_gpo_pol_ioctl)) > 0) {
			printk("cpx1024mon_ioctl: Bad copy in on set gpo pol ioctl\n");
			return -EFAULT;
		}

		for (gpo = 0; gpo < CPX1024_MAX_GPO + 1; gpo++) {
			if ((pol_ioctl.polarity[gpo] < 0) || (pol_ioctl.polarity[gpo] > CPX1024_MAX_GPO))
				printk("cpx1024mon_ioctl: invalid GPO number\n");
				return -EINVAL;
		}

		save_flags(flags);
		cli();

		for (gpo = 0; gpo < CPX1024_MAX_GPO + 1; gpo++) {
			cpx1024mon_regwrite(DTIMON_CMD_SET_GPOPOL, gpo, pol_ioctl.polarity[gpo], 0, 2);
		}

		restore_flags(flags);

		break;

	case CPX1024MON_GET_GPO_POL:
		save_flags(flags);
		cli();

		for (gpo = 0; gpo < CPX1024_MAX_GPO + 1; gpo++) {
			cpx1024mon_regread2(DTIMON_CMD_GET_GPOPOL, gpo, &pol_ioctl.polarity[gpo], 0, 1);
		}

		restore_flags(flags);

		/* Copy the information back to user space */
		if(copy_to_user((void *)arg,(const void *)&pol_ioctl,
				sizeof(struct cpx1024mon_gpo_pol_ioctl)) > 0) {
			printk("cpx1024mon_ioctl: Bad copy out on get gpo polarity ioctl\n");
			return -EFAULT;
		}

		break;

	case CPX1024MON_SET_GPI_POL:
		if (copy_from_user((void *)&ipol_ioctl, (const void *)arg,
				   sizeof(struct cpx1024mon_gpi_pol_ioctl)) > 0) {
			printk("cpx1024mon_ioctl: Bad copy in on set gpi pol ioctl\n");
			return -EFAULT;
		}

		save_flags(flags);
		cli();

		for (gpi = 0; gpi < CPX1024_MAX_GPI + 1; gpi++) {
			if ((ipol_ioctl.polarity[gpi] < 0) || (ipol_ioctl.polarity[gpi] > CPX1024_MAX_GPI_POL)) {
				restore_flags(flags);
				printk("cpx1024mon_ioctl: invalid GPI polarity\n");
				return -EINVAL;
			}
		}

		for (gpi = 0; gpi < CPX1024_MAX_GPI + 1; gpi++) {
			cpx1024mon_regwrite(DTIMON_CMD_SET_GPIPOL, gpi, ipol_ioctl.polarity[gpi], 0, 2);
		}

		restore_flags(flags);

		break;

	case CPX1024MON_GET_GPI_POL:
		save_flags(flags);
		cli();

		for (gpi = 0; gpi < CPX1024_MAX_GPI + 1; gpi++) {
			cpx1024mon_regread2(DTIMON_CMD_GET_GPIPOL, gpi, &ipol_ioctl.polarity[gpi], 0, 1);

			/* Enforce disabled as value 2 */
			if ((ipol_ioctl.polarity[gpi] != CPX1024MON_GPI_POL_LOW) &&
					(ipol_ioctl.polarity[gpi] != CPX1024MON_GPI_POL_HIGH)) {
				ipol_ioctl.polarity[gpi] = CPX1024MON_GPI_POL_DISABLE;
			}
		}

		restore_flags(flags);

		/* Copy the information back to user space */
		if(copy_to_user((void *)arg,(const void *)&ipol_ioctl,
				sizeof(struct cpx1024mon_gpi_pol_ioctl)) > 0) {
			printk("cpx1024mon_ioctl: Bad copy out on get gpi polarity ioctl\n");
			return -EFAULT;
		}

		break;

	case CPX1024MON_POWERDOWN:
		save_flags(flags);
		cli();

		cpx1024mon_regwrite(DTIMON_CMD_POWERDOWN, 0, 0, 0, 0);

		restore_flags(flags);

		break;

	case CPX1024MON_GET_VOLT:
		save_flags(flags);
		cli();

		for (volt = 0; volt < CPX1024_VOLT_SIZE; volt++) {
			cpx1024mon_regread2(DTIMON_CMD_GET_VOLT, volt,
					    &volt_ioctl.volt[volt][0], &volt_ioctl.volt[volt][1], 2);
		}

		restore_flags(flags);

		/* Copy the information back to user space */
		if(copy_to_user((void *)arg,(const void *)&volt_ioctl,
				sizeof(struct cpx1024mon_volt_ioctl)) > 0) {
			printk("cpx1024mon_ioctl: Bad copy out on get volt ioctl\n");
			return -EFAULT;
		}

		break;

	case CPX1024MON_GET_VTOLER:
		save_flags(flags);
		cli();

		for (volt = 0; volt < CPX1024_VOLT_TOL_SIZE; volt++) {
			cpx1024mon_regread2(DTIMON_CMD_GET_VTOLER, volt,
					    &vtoler_ioctl.volt[volt], NULL, 1);
		}

		restore_flags(flags);

		/* Copy the information back to user space */
		if(copy_to_user((void *)arg,(const void *)&vtoler_ioctl,
				sizeof(struct cpx1024mon_vtoler_ioctl)) > 0) {
			printk("cpx1024mon_ioctl: Bad copy out on get volt tolerance ioctl\n");
			return -EFAULT;
		}

		break;

	case CPX1024MON_SET_VTOLER:
		if (copy_from_user((void *)&vtoler_ioctl, (const void *)arg,
				   sizeof(struct cpx1024mon_vtoler_ioctl)) > 0) {
			printk("cpx1024mon_ioctl: Bad copy in on set volt tolerance ioctl\n");
			return -EFAULT;
		}

		save_flags(flags);
		cli();

		for (volt = 0; volt < CPX1024_VOLT_TOL_SIZE; volt++) {
			if ((vtoler_ioctl.volt[volt] < 0) || (vtoler_ioctl.volt[volt] > 100)) {
				restore_flags(flags);
				printk("cpx1024mon_ioctl: set voltages found tolerance of %d out of range\n", vtoler_ioctl.volt[volt]);
				return -EINVAL;
			}
		}

		for (volt = 0; volt < CPX1024_VOLT_TOL_SIZE; volt++) {
			cpx1024mon_regwrite(DTIMON_CMD_SET_VTOLER, volt, vtoler_ioctl.volt[volt], 0, 2);
		}

		restore_flags(flags);

		break;

	case CPX1024MON_GET_AMPS:
		save_flags(flags);
		cli();

		for (amp = 0; amp < CPX1024_AMP_SIZE; amp++) {
			cpx1024mon_regread2(DTIMON_CMD_GET_AMPS, amp,
					    &amp_ioctl.line[amp][0], &amp_ioctl.line[amp][1], 2);
		}

		restore_flags(flags);

		/* Copy the information back to user space */
		if(copy_to_user((void *)arg,(const void *)&amp_ioctl,
				sizeof(struct cpx1024mon_amp_ioctl)) > 0) {
			printk("cpx1024mon_ioctl: Bad copy out on get amps ioctl\n");
			return -EFAULT;
		}

		break;

	case CPX1024MON_GET_FAN_INFO:
		save_flags(flags);
		cli();

		for (fan = 0; fan < CPX1024_MAX_FAN; fan++) {
			cpx1024mon_regread2(DTIMON_CMD_GET_FANMODE, fan,
					    &fan_ioctl.fan[fan].mode, NULL, 1);
			cpx1024mon_regread2(DTIMON_CMD_GET_FANSTAT, fan,
					    &fan_ioctl.fan[fan].status, NULL, 1);
			cpx1024mon_regread2(DTIMON_CMD_GET_MINRPS, fan,
					    &fan_ioctl.fan[fan].rps_level, NULL, 1);
		}

		restore_flags(flags);

		/* Copy the information back to user space */
		if(copy_to_user((void *)arg,(const void *)&fan_ioctl,
				sizeof(struct cpx1024mon_fan_ioctl)) > 0) {
			printk("cpx1024mon_ioctl: Bad copy out on get fan info ioctl\n");
			return -EFAULT;
		}

		break;

	case CPX1024MON_SET_FAN_INFO:
		if (copy_from_user((void *)&fan_ioctl, (const void *)arg,
				   sizeof(struct cpx1024mon_fan_ioctl)) > 0) {
			printk("cpx1024mon_ioctl: Bad copy in on set fan info ioctl\n");
			return -EFAULT;
		}

		for (fan = 0; fan < CPX1024_MAX_FAN; fan++) {
			switch (fan_ioctl.fan[fan].mode) {
			case DTIMON_GSFAN_STROBE:
			case DTIMON_GSFAN_LEVEL:
			case DTIMON_GSFAN_DISABLE:
				break;
			default:
				printk("cpx1024mon_ioctl: invalid fan mode for fan %d\n", fan);
				return -EINVAL;
			}
		}

		save_flags(flags);
		cli();

		for (fan = 0; fan < CPX1024_MAX_FAN; fan++) {
			cpx1024mon_regwrite(DTIMON_CMD_SET_FANMODE, fan, fan_ioctl.fan[fan].mode, 0, 2);
			if (fan_ioctl.fan[fan].mode != DTIMON_GSFAN_DISABLE) {
				cpx1024mon_regwrite(DTIMON_CMD_SET_MINRPS, fan,
						    fan_ioctl.fan[fan].rps_level, 0, 2);
			}
			cpx1024mon_regread2(DTIMON_CMD_GET_FANSTAT, fan,
					    &fan_ioctl.fan[fan].status, NULL, 1);
		}

		restore_flags(flags);

		/* Copy the information back to user space */
		if(copy_to_user((void *)arg,(const void *)&fan_ioctl,
				sizeof(struct cpx1024mon_fan_ioctl)) > 0) {
			printk("cpx1024mon_ioctl: Bad copy out on get fan info ioctl\n");
			return -EFAULT;
		}

		break;

	case CPX1024MON_GET_TEMPMASK:
		save_flags(flags);
		cli();

		cpx1024mon_regread2(DTIMON_CMD_GET_TIBM, 0, &tibm0, 0, 1);
		cpx1024mon_regread2(DTIMON_CMD_GET_TIBM, 1, &tibm1, 0, 1);
		tibm0 += tibm1 << 8;

		restore_flags(flags);

		/* Copy the information back to user space */
		if(copy_to_user((void *)arg,(const void *)&tibm0, sizeof(int)) > 0) {
			printk("cpx1024mon_ioctl: Bad copy out on get temp install bit mask ioctl\n");
			return -EFAULT;
		}

		break;

	case CPX1024MON_GET_CPUS:
		save_flags(flags);
		cli();

		cpx1024mon_regread2(DTIMON_CMD_GET_CPUMASK, 0, &cpumask, 0, 1);

		restore_flags(flags);

		/* Copy the information back to user space */
		if(copy_to_user((void *)arg,(const void *)&cpumask, sizeof(int)) > 0) {
			printk("cpx1024mon_ioctl: Bad copy out on get temp install bit mask ioctl\n");
			return -EFAULT;
		}

		break;

	case CPX1024MON_GET_MON_INFO:
		save_flags(flags);
		cli();

		cpx1024mon_regread(DTIMON_CMD_GET_BIOSREV, &minfo_ioctl.bios_rev, NULL, 1);
		cpx1024mon_regread(DTIMON_CMD_GET_MONREV, &minfo_ioctl.mon_rev, &minfo_ioctl.mon_patch, 2);

		restore_flags(flags);

		/* Copy the information back to user space */
		if(copy_to_user((void *)arg,(const void *)&minfo_ioctl,
				sizeof(struct cpx1024mon_info)) > 0) {
			printk("cpx1024mon_ioctl: Bad copy out on get mon info ioctl\n");
			return -EFAULT;
		}

		break;

	case CPX1024MON_UPDATE_EEPROM:
		save_flags(flags);
		cli();

		cpx1024mon_regwrite(DTIMON_CMD_SET_EEPROMM, 0, 0, 0, 0);

		restore_flags(flags);

		break;

	case CPX1024MON_SERIAL_TRANSFER:
		if (copy_from_user((void *)&transfer, (const void *)arg, sizeof(int)) > 0) {
			printk("cpx1024mon_ioctl: Bad copy in on term transfer ioctl\n");
			return -EFAULT;
		}

		save_flags(flags);
		cli();

		if (transfer) {
			cpx1024mon_regwrite(DTIMON_CMD_ENABLE_TTM, 0, 0, 0, 0);
		} else {
			cpx1024mon_regwrite(DTIMON_CMD_DISABLE_TTM, 0, 0, 0, 0);
		}

		restore_flags(flags);

		break;

	case CPX1024MON_SERIAL_VERBOSE:
		if (copy_from_user((void *)&verbose, (const void *)arg, sizeof(int)) > 0) {
			printk("cpx1024mon_ioctl: Bad copy in on serial verbose ioctl\n");
			return -EFAULT;
		}

		save_flags(flags);
		cli();

		if (verbose) {
			cpx1024mon_regwrite(DTIMON_CMD_VERBOSE, DTIMON_VERBOSE_ON, 0, 0, 1);
		} else {
			cpx1024mon_regwrite(DTIMON_CMD_VERBOSE, DTIMON_VERBOSE_OFF, 0, 0, 1);
		}
		
		restore_flags(flags);

		break;

	case CPX1024MON_SERIAL_PASSWD:
		if (copy_from_user((void *)&passwd_ioctl, (const void *)arg,
				   sizeof(struct cpx1024mon_passwd_ioctl)) > 0) {
			printk("cpx1024mon_ioctl: Bad copy in on set passwd ioctl\n");
			return -EFAULT;
		}

		save_flags(flags);
		cli();

		cpx1024mon_regwrite(DTIMON_CMD_SET_PASSWD, passwd_ioctl.pw[0],
					    passwd_ioctl.pw[1], passwd_ioctl.pw[2], 1);

		restore_flags(flags);

		break;

	case CPX1024MON_SERIAL_SETUID:
		if (copy_from_user((void *)&sysuid_ioctl, (const void *)arg,
				   sizeof(struct cpx1024mon_passwd_ioctl)) > 0) {
			printk("cpx1024mon_ioctl: Bad copy in on set uid ioctl\n");
			return -EFAULT;
		}

		save_flags(flags);
		cli();

		cpx1025mon_setuid(sysuid_ioctl.uid);

		restore_flags(flags);

		break;

	case CPX1024MON_SERIAL_ECHO:
		if (copy_from_user((void *)&echo_mode, (const void *)arg, sizeof(int)) > 0) {
			printk("cpx1024mon_ioctl: Bad copy in on serial echo ioctl\n");
			return -EFAULT;
		}

		save_flags(flags);
		cli();

		if (echo_mode) {
			cpx1024mon_regwrite(DTIMON_CMD_SERIAL_ECHO, DTIMON_ECHO_ON, 0, 0, 1);
		} else {
			cpx1024mon_regwrite(DTIMON_CMD_SERIAL_ECHO, DTIMON_ECHO_OFF, 0, 0, 1);
		}

		restore_flags(flags);

		break;

	case CPX1024MON_REG_EVENTS:
		/* Access to the eventing mechanism for this driver is controlled
		 * on a thread basis.  Each thread registers to use the eventing
		 * mechanism and gets a queue assigned to it.  Access to the queue
		 * (see below) is controlled by checking the pid of the thread to
		 * see if it is registered.
		 */
		save_flags(flags);
		cli();

		/* Make sure the thread is not currently registered */
		tmpUserPtr = cpx1024monUserList;
		while (tmpUserPtr) {
			if (tmpUserPtr->pid == current->pid) {
				restore_flags(flags);
				printk("CPX1024MON_REG_EVENTS: One resister per process\n");
				return -EEXIST;
			}
			tmpUserPtr = tmpUserPtr->next;
		}

		/* Is it possible to register another user? */
		if (cpx1024monUserFreeList == (struct cpx1024mon_user *)NULL) {
			printk("CPX1024MON_REG_EVENTS: Already at max usage\n");
			restore_flags(flags);
			return -ENOMEM;
		}

		/* Get a user control structure */
		tmpUserPtr = cpx1024monUserFreeList;
		cpx1024monUserFreeList = cpx1024monUserFreeList->next;

		tmpUserPtr->pid = current->pid;
		tmpUserPtr->head = (struct cpx1024mon_message *)NULL;
		tmpUserPtr->tail = (struct cpx1024mon_message *)NULL;
  
		/* Add it to the list of users */
		tmpUserPtr->next = cpx1024monUserList;
		cpx1024monUserList = tmpUserPtr;

		restore_flags(flags);

		break;

	case CPX1024MON_UNREG_EVENTS:
		/* Unregister a thread from the eventing mechanism */
		save_flags(flags);
		cli();

		delUserPtr = (struct cpx1024mon_user *)NULL;

		/* Find the User entry to unregister */
		tmpUserPtr = cpx1024monUserList;
		
		if (tmpUserPtr == (struct cpx1024mon_user *)NULL) {
			break;
		}
		
		if (tmpUserPtr->pid == current->pid) { /* Head is special */
			cpx1024monUserList = cpx1024monUserList->next;
			delUserPtr = tmpUserPtr;
		} else {
			while (tmpUserPtr->next) {
				if (tmpUserPtr->next->pid == current->pid) {
					delUserPtr = tmpUserPtr->next;
					tmpUserPtr->next = delUserPtr->next;
				}
				if (tmpUserPtr->next) {
					tmpUserPtr = tmpUserPtr->next;
				}
			}
		}
			
		if (delUserPtr) {
				/* Free any pending messages */
			tmpMsgPtr = delUserPtr->head;
			while (tmpMsgPtr) {
				freeMsgPtr = tmpMsgPtr;
				tmpMsgPtr = tmpMsgPtr->next;
				freeMsgPtr->next = cpx1024monMessFreeList;
				cpx1024monMessFreeList = freeMsgPtr;
				freeMsgPtr->code = 0;
			}
			
			delUserPtr->next = cpx1024monUserFreeList;
			cpx1024monUserFreeList = delUserPtr;
		}
		
		restore_flags(flags);

		break;

	case CPX1024MON_SEND_ALL_EVENTS:
		/* This ioctl checks all the possible events from the hardware
		 * monitor and queues them through the normal eventing mechanism.
		 */
		save_flags(flags);
		cli();
		cpx1024mon_send_all_event();
		restore_flags(flags);
		break;

	case CPX1024MON_GET_EVENTS:
		/* Get current messages (up to 10 at a time) waiting for the user */
		if (copy_from_user((void *)&msg_buf, (const void *)arg,
				   sizeof(struct cpx1024mon_msg_ioctl)) > 0) {
			printk("cpx1024mon_ioctl: Bad copy in on ioctl read message\n");
			return -EFAULT;
		}

		/* Get the user pointer */
		save_flags(flags);
		cli();
		tmpUserPtr = cpx1024monUserList;
		while (tmpUserPtr) {
			if (tmpUserPtr->pid == current->pid) {
				break;
			}
			tmpUserPtr = tmpUserPtr->next;
		}

		if (tmpUserPtr->pid != current->pid) {
			restore_flags(flags);
			printk("cpx1024mon_ioctl: bad user juju pids %d and %d\n",
			       tmpUserPtr->pid, current->pid);
			return -EIO;
		}

		/* Ensure we do not over shoot the end of the structure */
		if (msg_buf.count > CPX1024_MAX_MSG)
			 msg_buf.count = CPX1024_MAX_MSG;

		if (msg_buf.count < 0) {
			restore_flags(flags);
			printk("cpx1024mon_ioctl: CPX1024MON_GET_EVENTS illegal count %d\n", msg_buf.count);
			return -EINVAL;
		}

		/* If there is currently nothing to wait for then sleep */
		if (tmpUserPtr->head == (struct cpx1024mon_message *)NULL) {
			restore_flags(flags);
			interruptible_sleep_on(&tmpUserPtr->wq);
			save_flags(flags);
			cli();
		}

		/* Copy the messages to the return structure */
		tmpMsgPtr = tmpUserPtr->head;
		while ((msg_num < msg_buf.count) && (tmpMsgPtr != (struct cpx1024mon_message *)NULL)) {
			msg_buf.msgs[msg_num].code = tmpMsgPtr->code;
			msg_buf.msgs[msg_num].val = tmpMsgPtr->val;

			/* Free the message pointer */
			freeMsgPtr = tmpMsgPtr;
			tmpMsgPtr = tmpMsgPtr->next;
			freeMsgPtr->next = cpx1024monMessFreeList;
			cpx1024monMessFreeList = freeMsgPtr;
			freeMsgPtr->code = 0;

			msg_num++;
		}

		/* Update the count to reflect the actual number sent */
		msg_buf.count = msg_num;

		/* Update the message pointer */
		if (tmpMsgPtr != (struct cpx1024mon_message *)NULL) {
			tmpUserPtr->head = tmpMsgPtr;
		} else {
			tmpUserPtr->head = NULL;
			tmpUserPtr->tail = NULL;
		}

		restore_flags(flags);

		/* Copy the information back to user space */
		if(copy_to_user((void *)arg,(const void *)&msg_buf,
				sizeof(struct cpx1024mon_msg_ioctl)) > 0) {
			printk("cpx1024mon_ioctl: Bad copy out on ioctl read message\n");
			return -EFAULT;
		}

		break;
	default:
		return -EINVAL;
		break;
	}

	return 0;
}

/*
 * Linux open entry point.
 */
static int
cpx1024mon_open(struct inode *inode, struct file *file)
{
	cpx1024mon_user_cnt++;

	return 0;
}

static int
cpx1024mon_release(struct inode *inode, struct file *file)
{
	struct cpx1024mon_user *tmpUserPtr;
	struct cpx1024mon_user *delUserPtr = (struct cpx1024mon_user *)NULL;
	struct cpx1024mon_message *tmpMsgPtr;
	struct cpx1024mon_message *delMsgPtr;

	/* If no users registered at all then make sure the
	 * list gets cleared out.  This is a simple sanity check.
	 */
	if (--cpx1024mon_user_cnt == 0) {
		tmpUserPtr = cpx1024monUserList;
		while (tmpUserPtr) {
			cpx1024monUserList = tmpUserPtr->next;

			/* Free any pending messages */
			tmpMsgPtr = tmpUserPtr->head;
			while (tmpMsgPtr) {
				delMsgPtr = tmpMsgPtr;
				tmpMsgPtr = tmpMsgPtr->next;
				delMsgPtr->next = cpx1024monMessFreeList;
				cpx1024monMessFreeList = delMsgPtr;
				delMsgPtr->code = 0;
			}

			tmpUserPtr->next = cpx1024monUserFreeList;
			cpx1024monUserFreeList = tmpUserPtr;

			tmpUserPtr = cpx1024monUserList;
		}

		return 0;
	}

	/* Find the user structure */
	tmpUserPtr = cpx1024monUserList;

	if (tmpUserPtr == (struct cpx1024mon_user *)NULL) {
		return 0;
	}

	if (tmpUserPtr->pid == current->pid) { /* Head is special */
		cpx1024monUserList = cpx1024monUserList->next;
		delUserPtr = tmpUserPtr;
		
	} else {
		while (tmpUserPtr->next) {
			if (tmpUserPtr->next->pid == current->pid) {
				delUserPtr = tmpUserPtr->next;
				tmpUserPtr->next = delUserPtr->next;
			}
			if (tmpUserPtr->next) {
				tmpUserPtr = tmpUserPtr->next;
			}
		}
	}

	if (delUserPtr == (struct cpx1024mon_user *)NULL) {
		return 0;
	}

	/* Free any pending messages */
	tmpMsgPtr = delUserPtr->head;
	while (tmpMsgPtr) {
		delMsgPtr = tmpMsgPtr;
		tmpMsgPtr = tmpMsgPtr->next;
		delMsgPtr->next = cpx1024monMessFreeList;
		cpx1024monMessFreeList = delMsgPtr;
		delMsgPtr->code = 0;
		
	}

#ifdef OMIT
	/* Free the user control structure */
	delUserPtr->next = cpx1024monUserFreeList;
	cpx1024monUserFreeList = delUserPtr;
	delUserPtr->pid = 0;
	delUserPtr->head = (struct cpx1024mon_message *)NULL;
	delUserPtr->tail = (struct cpx1024mon_message *)NULL;
#endif

	return 0;
}

/*
 * These variables keep track of the last known value for all
 * status bits in the hardware monitor.  The poll routine must
 * know this to test if the bit has changed or not.
 */
static int offbt_bool = 0;
static int fan_error_mask = 0;
static int gpi_error_mask = 0;
static int cpu_was_on = 0;
static int minor_was_on = 0;
static int major_was_on = 0;
static int temp_mask[2] = {0, 0};
static int volt_mask[2] = {0, 0};  /* Mask for each cpu */

/*
 * The bulk of the eventing is done by this poll routine.  It
 * runs every second or so checking the status of all alarms
 * in the hardware monitor
 */
void
cpx1024mon_poll(unsigned long nothing)
{
	int status;
	int sys_status;
	int fan_val;
	int fan;
	int gpi_val;
	int gpi;
	int volt;
	int temp;
	int volt_vals[2];
	int temp_vals[2];
	int temp_return;
	int volt_return;
	unsigned long flags;

	/* Read the status from the monitor */
	status = inb(DTIMON_CMD);
	if (status == 0) {
		/* Reset the timeout function for the next poll */
		cpx1024mon_timer.function = cpx1024mon_poll;
		cpx1024mon_timer.data = (unsigned long)NULL;
		cpx1024mon_timer.expires = jiffies + cpx1024mon_timeout_interval * HZ;
		add_timer(&cpx1024mon_timer);

		return;
	}

	/* This is where a normal interrupt routine starts */
	if (status & DTIMON_STATUS_INTP) {
		cpx1024mon_regread(DTIMON_CMD_GET_STATUS, &sys_status, NULL, 1);
	}

	save_flags(flags);
	cli();
	
	if (sys_status & DTIMON_STATUS_VOLT || volt_mask[0] || volt_mask[1]) {  /* Voltage alarm */
		cpx1024mon_regread(DTIMON_CMD_GET_VOLTAGE, &volt_vals[0], &volt_vals[1], 2);
		for (volt = 0; volt < DTIMON_VOLT_BYTE1_MAX; volt++) {
			if (volt_vals[0] & (0x1 << volt)) {
				if (!(volt_mask[0] & (0x1 << volt))) {
					cpx1024mon_sendevent(0, DTIMON_ALRM_VOLT_ON, volt);
				}
			} else {
				if (volt_mask[0] & (0x1 << volt)) {
					cpx1024mon_sendevent(0, DTIMON_ALRM_VOLT_OFF, volt);
				}
			}
		}

		volt_mask[0] = volt_vals[0];

		for (volt = 0; volt < DTIMON_VOLT_BYTE2_MAX; volt++) {
			if (volt_vals[1] & (0x1 << volt)) {
				if (!(volt_mask[1] & (0x1 << volt))) {
					volt_return = volt + DTIMON_VOLT_BYTE1_MAX;
					cpx1024mon_sendevent(0, DTIMON_ALRM_VOLT_ON, volt_return);
				}
			} else {
				if (volt_mask[1] & (0x1 << volt)) {
					volt_return = volt + DTIMON_VOLT_BYTE1_MAX;
					cpx1024mon_sendevent(0, DTIMON_ALRM_VOLT_OFF, volt_return);
				}
			}
		}

		volt_mask[1] = volt_vals[1];
	}

	if (sys_status & DTIMON_STATUS_ONBT || temp_mask[0] || temp_mask[1]) { /* On board temp */
		cpx1024mon_regread(DTIMON_CMD_GET_TEMP, &temp_vals[0], &temp_vals[1], 2);
		for (temp = 0; temp < DTIMON_TEMP_BYTE1_MAX; temp++) {
			if (temp_vals[0] & (0x1 << temp)) {
				if (!(temp_mask[0] & (0x1 << temp))) {
					cpx1024mon_sendevent(0, DTIMON_ALRM_ONBTEMP_ON, temp);
				}
			} else {
				if (temp_mask[0] & (0x1 << temp)) {
					cpx1024mon_sendevent(0, DTIMON_ALRM_ONBTEMP_OFF, temp);
				}
			}
		}

		temp_mask[0] = temp_vals[0];

		for (temp = 0; temp < DTIMON_TEMP_BYTE2_MAX; temp++) {
			if (temp_vals[1] & (0x1 << temp)) {
				if (!(temp_mask[1] & (0x1 << temp))) {
					temp_return = temp + DTIMON_TEMP_BYTE1_MAX;
					cpx1024mon_sendevent(0, DTIMON_ALRM_ONBTEMP_ON, temp_return);
				}
			} else {
				if (temp_mask[1] & (0x1 << temp)) {
					temp_return = temp + DTIMON_TEMP_BYTE1_MAX;
					cpx1024mon_sendevent(0, DTIMON_ALRM_ONBTEMP_OFF, temp_return);
				}
			}
		}

		temp_mask[1] = temp_vals[1];
	}

	if (sys_status & DTIMON_STATUS_OFFBT || offbt_bool) {      /* Off board temp  alarm */
		if (sys_status & DTIMON_STATUS_OFFBT) {
			if(!offbt_bool) {  /* New alarm */
				cpx1024mon_sendevent(0, DTIMON_ALRM_OFFBTEMP_ON, 0);
			}
		} else {
			if (offbt_bool) {
				cpx1024mon_sendevent(0, DTIMON_ALRM_OFFBTEMP_OFF, 0);
			}
		}
	}

	if ((sys_status & DTIMON_STATUS_FANS) || fan_error_mask) { /* Fan alarm changed */
		cpx1024mon_regread(DTIMON_CMD_GET_FAN, &fan_val, NULL, 1);

		for (fan = 1; fan <= DTIMON_FAN_MAX; fan++) {
			if (fan_val & (0x1 << (fan - 1))) {  /* Status of this fan if failed */
				if (!(fan_error_mask & (0x1 << (fan - 1)))) { /* New failure */
					/* Fan alarm changed to on */
					cpx1024mon_sendevent(0, DTIMON_ALRM_FAN_ON, fan);
				}
			} else {
				if (fan_error_mask & (0x1 << (fan - 1))) { /* Reset failure */
					/* Fan alarm changed to off */
					cpx1024mon_sendevent(0, DTIMON_ALRM_FAN_OFF, fan);
				}
			}
		}

		fan_error_mask = fan_val;
	}

	if (sys_status & DTIMON_STATUS_GPI || gpi_error_mask) {   /* General purpose input alarm */
		cpx1024mon_regread(DTIMON_CMD_GET_GPI, &gpi_val, NULL, 1);

		for (gpi = 1; gpi <= DTIMON_FAN_MAX; gpi++) {
			if (gpi_val & (0x1 << (gpi - 1))) {  /* Status of this gpi is failed */
				if (!(gpi_error_mask & (0x1 << (gpi - 1)))) { /* New failure */
					/* GPI alarm changed to on */
					cpx1024mon_sendevent(0, DTIMON_ALRM_GPI_ON, gpi);
				}
			} else {
				if (gpi_error_mask & (0x1 << (gpi - 1))) { /* Reset failure */
					/* GPI alarm changed to off */
					cpx1024mon_sendevent(0, DTIMON_ALRM_GPI_OFF, gpi);
				}
			}
		}

		gpi_error_mask = gpi_val;
	}

	if ((sys_status & DTIMON_STATUS_CPU) || cpu_was_on) {  /* CPU or Power module failure alarm */
		if (sys_status & DTIMON_STATUS_CPU) {
			if (!cpu_was_on) {
				cpx1024mon_sendevent(0, DTIMON_ALRM_CPU_ON, 0);
				cpu_was_on = 1;
			}
		} else {
			cpx1024mon_sendevent(0, DTIMON_ALRM_CPU_OFF, 0);
			cpu_was_on = 0;
		}

		cpu_was_on = sys_status & DTIMON_STATUS_CPU;
	}

	if ((sys_status & DTIMON_STATUS_MINOR) || minor_was_on) { /* Minor indicator lit */
		if (sys_status & DTIMON_STATUS_MINOR) {
			if (!minor_was_on) {
				cpx1024mon_sendevent(0, DTIMON_ALRM_MINORLED_ON, 0);	
				minor_was_on = 1;
			}
		} else {
			cpx1024mon_sendevent(0, DTIMON_ALRM_MINORLED_OFF, 0);
			minor_was_on = 0;
		}
		minor_was_on = sys_status & DTIMON_STATUS_MINOR;
	}

	if ((sys_status & DTIMON_STATUS_MAJOR) || major_was_on) { /* Major indicator lit */
		if (sys_status & DTIMON_STATUS_MAJOR) {
			if (!major_was_on) {
				cpx1024mon_sendevent(0, DTIMON_ALRM_MAJORLED_ON, 0);
				major_was_on = 1;
			}
		} else {
			cpx1024mon_sendevent(0, DTIMON_ALRM_MAJORLED_OFF, 0);
			major_was_on = 0;
		}
		major_was_on = sys_status & DTIMON_STATUS_MAJOR;
	}

	restore_flags(flags);

	/* Reset the timeout function for the next poll */
	cpx1024mon_timer.function = cpx1024mon_poll;
	cpx1024mon_timer.data = (unsigned long)NULL;
	cpx1024mon_timer.expires = jiffies + cpx1024mon_timeout_interval * HZ;
	add_timer(&cpx1024mon_timer);

	return;
}

/*
 * Check all status bits and queue the appropriate event.
 */
void
cpx1024mon_send_all_event(void)
{
	int volt;
	int temp;
	int fan;
	int gpi;
	int temp_return;
	int volt_return;
	
	for (volt = 0; volt < DTIMON_VOLT_BYTE1_MAX; volt++) {
		if (volt_mask[0] & (0x1 << volt)) {
			cpx1024mon_sendevent(current->pid, DTIMON_ALRM_VOLT_ON, volt);
		} else {
			cpx1024mon_sendevent(current->pid, DTIMON_ALRM_VOLT_OFF, volt);
		}
	}
	
	for (volt = 0; volt < DTIMON_VOLT_BYTE2_MAX; volt++) {
		if (volt_mask[1] & (0x1 << volt)) {
			volt_return = volt + DTIMON_VOLT_BYTE1_MAX;
			cpx1024mon_sendevent(current->pid, DTIMON_ALRM_VOLT_ON, volt_return);
		} else {
			volt_return = volt + DTIMON_VOLT_BYTE1_MAX;
			cpx1024mon_sendevent(current->pid, DTIMON_ALRM_VOLT_OFF, volt_return);
		}
	}

	for (temp = 0; temp < DTIMON_TEMP_BYTE1_MAX; temp++) {
		if (temp_mask[0] & (0x1 << temp)) {
			cpx1024mon_sendevent(current->pid, DTIMON_ALRM_ONBTEMP_ON, temp);
		} else {
			cpx1024mon_sendevent(current->pid, DTIMON_ALRM_ONBTEMP_OFF, temp);
		}
	}
	
	for (temp = 0; temp < DTIMON_TEMP_BYTE2_MAX; temp++) {
		if (temp_mask[1] & (0x1 << temp)) {
			temp_return = temp + DTIMON_TEMP_BYTE1_MAX;
			cpx1024mon_sendevent(current->pid, DTIMON_ALRM_ONBTEMP_ON, temp_return);
		} else {
			temp_return = temp + DTIMON_TEMP_BYTE1_MAX;
			cpx1024mon_sendevent(current->pid, DTIMON_ALRM_ONBTEMP_OFF, temp_return);
		}
		
	}
	
	if (offbt_bool) {
		cpx1024mon_sendevent(current->pid, DTIMON_ALRM_OFFBTEMP_ON, 0);
	} else {
		cpx1024mon_sendevent(current->pid, DTIMON_ALRM_OFFBTEMP_OFF, 0);
	}

	for (fan = 1; fan <= DTIMON_FAN_MAX; fan++) {
		if (fan_error_mask & (0x1 << (fan - 1))) {
			cpx1024mon_sendevent(current->pid, DTIMON_ALRM_FAN_ON, fan);
		} else {
			cpx1024mon_sendevent(current->pid, DTIMON_ALRM_FAN_OFF, fan);
		}
	}

	for (gpi = 1; gpi <= DTIMON_FAN_MAX; gpi++) {
		if (gpi_error_mask & (0x1 << (gpi - 1))) {
			cpx1024mon_sendevent(current->pid, DTIMON_ALRM_GPI_ON, gpi);
		} else {
			cpx1024mon_sendevent(current->pid, DTIMON_ALRM_GPI_OFF, gpi);
		}
	}

	if (cpu_was_on) {
		cpx1024mon_sendevent(current->pid, DTIMON_ALRM_CPU_ON, 0);
	} else {
		cpx1024mon_sendevent(current->pid, DTIMON_ALRM_CPU_OFF, 0);
	}

	if (minor_was_on) {
		cpx1024mon_sendevent(0, DTIMON_ALRM_MINORLED_ON, 0);
	} else {
		cpx1024mon_sendevent(0, DTIMON_ALRM_MINORLED_OFF, 0);
	}

	if (major_was_on) {
		cpx1024mon_sendevent(0, DTIMON_ALRM_MAJORLED_ON, 0);
	} else {
		cpx1024mon_sendevent(0, DTIMON_ALRM_MAJORLED_OFF, 0);
	}
}

/*
 * Read a value from the hardware monitor ISA interface.
 *
 * This function is for a monitor command that does not have a modifier.
 */
void
cpx1024mon_regread(unsigned char cmd, int * val1, int *val2, int num)
{
	unsigned char status = inb(DTIMON_CMD);
	int loop = 0;

	while (!(status & DTIMON_STATUS_CBE)) {
		status = inb(DTIMON_CMD);
	}

	outb(cmd, DTIMON_CMD);

	status = inb(DTIMON_CMD);
	while (!(status & DTIMON_STATUS_CBE)) {
		status = inb(DTIMON_CMD);
	}

	while (loop < num) {
		status = inb(DTIMON_CMD);
		while (!(status & DTIMON_STATUS_OBF)) {
			status = inb(DTIMON_CMD);
		}

		if (loop == 0)
			*val1 = (int)inb(DTIMON_DATA);
		else
			*val2 = (int)inb(DTIMON_DATA);

		loop++;
	}
}

/*
 * Read a value from the hardware monitor ISA interface.
 *
 * This function is for a monitor command with a modifier.
 */
void
cpx1024mon_regread2(unsigned char cmd, int mod, int * val1, int *val2, int num)
{
	unsigned char status = inb(DTIMON_CMD);
	int loop = 0;

	while (!(status & DTIMON_STATUS_CBE)) {
		status = inb(DTIMON_CMD);
	}

	outb(cmd, DTIMON_CMD);

	status = inb(DTIMON_CMD);
	while (!(status & DTIMON_STATUS_CBE)) {
		status = inb(DTIMON_CMD);
	}
	status = inb(DTIMON_CMD);
	while (!(status & DTIMON_STATUS_IBE)) {
		status = inb(DTIMON_CMD);
	}

	outb(mod, DTIMON_DATA);

	while (loop < num) {
		status = inb(DTIMON_CMD);
		while (!(status & DTIMON_STATUS_OBF)) {
			status = inb(DTIMON_CMD);
		}

		if (loop == 0)
			*val1 = (int)inb(DTIMON_DATA);
		else
			*val2 = (int)inb(DTIMON_DATA);

		loop++;
	}
}

/*
 * Write a command to the hardware monitor.
 */
void
cpx1024mon_regwrite(unsigned char cmd, int val1, int val2, int val3, int num)
{
	unsigned char status = inb(DTIMON_CMD);
	int loop = 0;

	while (!(status & DTIMON_STATUS_CBE)) {
		status = inb(DTIMON_CMD);
	}

	outb(cmd, DTIMON_CMD);

	status = inb(DTIMON_CMD);
	while (!(status & DTIMON_STATUS_CBE)) {
		status = inb(DTIMON_CMD);
	}

	while (loop < num) {
		status = inb(DTIMON_CMD);
		while (!(status & DTIMON_STATUS_IBE)) {
			status = inb(DTIMON_CMD);
		}

		if (loop == 0) {
			outb(val1, DTIMON_DATA);
 		} else if (loop == 1){
			outb(val2, DTIMON_DATA);
		} else {
			outb(val3, DTIMON_DATA);
		}

		loop++;
	}
}

/*
 * Write the unit ID to the hardware monitor.
 */
void
cpx1025mon_setuid(char *uid)
{
	unsigned char status = inb(DTIMON_CMD);
	int loop = 0;

	while (!(status & DTIMON_STATUS_CBE)) {
		status = inb(DTIMON_CMD);
	}

	outb(DTIMON_CMD_SET_UNITID, DTIMON_CMD);

	status = inb(DTIMON_CMD);
	while (!(status & DTIMON_STATUS_CBE)) {
		status = inb(DTIMON_CMD);
	}

	for (loop = 0; loop < 10; loop++) {
		status = inb(DTIMON_CMD);
		while (!(status & DTIMON_STATUS_IBE)) {
			status = inb(DTIMON_CMD);
		}

		if (*uid == (char)0) {
			outb((char)0x2a, DTIMON_DATA);
			return;
		}

		outb(*uid++, DTIMON_DATA);
	}
}

/*
 * Allocate a messages structure from the free queue.  The init routine
 * allocates a chunk of memory and formats it.  Under high load this may
 * not be enough so this routine, when it finds an empty free list,
 * allocates more memory and adds it to the free list.  This memory is
 * never returned to the system.
 */
struct cpx1024mon_message *
cpx1024mon_allocMsg(void)
{
	struct cpx1024mon_message *tmpMsgPtr;
	struct cpx1024mon_message *tmpMessPtr;
	int loop;

	tmpMsgPtr = cpx1024monMessFreeList;
	cpx1024monMessFreeList = cpx1024monMessFreeList->next;
	tmpMsgPtr->next = (struct cpx1024mon_message *)NULL;

	if (cpx1024monMessFreeList == (struct cpx1024mon_message *)NULL) {
		cpx1024monMessFreeList = kmalloc(CPX1024_MESSAGE_CHUNK_SIZE *
						 sizeof(struct cpx1024mon_message), GFP_ATOMIC);

		tmpMessPtr = cpx1024monMessFreeList;

		for (loop = 0; loop < CPX1024_MESSAGE_CHUNK_SIZE - 1; loop++) {
			tmpMessPtr->code = 0;
			tmpMessPtr->next = tmpMessPtr + 1;
			tmpMessPtr++;
		}
		tmpMessPtr->code = 0;
		tmpMessPtr->next = (struct cpx1024mon_message *)NULL;	
	}

	return tmpMsgPtr;
}

/*
 * Queue an event to all registered users.
 */
void
cpx1024mon_sendevent(int pid, int ret_code, int val)
{
	struct cpx1024mon_user *tmpUserPtr;

	/* Get the user pointer */
	tmpUserPtr = cpx1024monUserList;
	while (tmpUserPtr) {
		if (pid && (tmpUserPtr->pid != pid)) {
			tmpUserPtr = tmpUserPtr->next;
			continue;
		}

		if (tmpUserPtr->head == (struct cpx1024mon_message *)NULL) {
			tmpUserPtr->head = cpx1024mon_allocMsg();
			tmpUserPtr->tail = tmpUserPtr->head;
		} else {
			tmpUserPtr->tail->next = cpx1024mon_allocMsg();
			tmpUserPtr->tail = tmpUserPtr->tail->next;
		}

		/* Fill in the information to be sent back */
		tmpUserPtr->tail->code = ret_code;
		tmpUserPtr->tail->val = val;

		/* Wake waiting process */
		wake_up_interruptible(&tmpUserPtr->wq);
		tmpUserPtr = tmpUserPtr->next;
	}
}

/*
 * Local variables:
 *  c-indent-level: 8
 *  c-basic-offset: 8
 *  tab-width: 8
 * End:
 */
