/*
 * $Id: physmap.c,v 1.10 2001/03/03 18:17:09 dwmw2 Exp $
 *
 * Normal mappings of chips in physical memory
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/config.h>
#include <linux/mtd/partitions.h>

static int probe_done = 0;
#define WINDOW_ADDR         0xbf800000
#define WINDOW_SIZE         0x800000

#define BUSWIDTH CONFIG_MTD_PHYSMAP_BUSWIDTH

static struct mtd_info *mymtd;

#define BOOTL_PART_SIZE                     0xA0000 /* 640 k */
#define DUPLICATED_BOOTL_PART_SIZE          0x10000 /* 64 k - -space for duplicating first 
                                                      statge and functional keys only */
#define JFFS2_PART_SIZE                     (WINDOW_SIZE - BOOTL_PART_SIZE - DUPLICATED_BOOTL_PART_SIZE)

#define BOOTL_PART_STARTING_OFFSET          0x00000
#define JFFS2_PART_STARTING_OFFSET          0xA0000 
#define DUPLCATED_BOOT_PART_STARTING_OFFSET (WINDOW_SIZE - 0x10000)
#define DUPLCATED_BOOT_PART_PHY_START_OFFSET (WINDOW_SIZE/2)

static struct mtd_partition myboard_partitions[] = {
    {
        name: "bootstrap and kernel",
        size:  BOOTL_PART_SIZE,  /* 640k */
        offset: BOOTL_PART_STARTING_OFFSET,
    },
    {
        name: "jffs2 filesystem",
        size: JFFS2_PART_SIZE, /* 8M - 640 k - 64k*/
        offset: JFFS2_PART_STARTING_OFFSET,
    },
    {
        name: "duplicated bootloader",
        size: DUPLICATED_BOOTL_PART_SIZE, /* 64k*/
        offset:DUPLCATED_BOOT_PART_STARTING_OFFSET,
    },
};

#define NUM_PARTITIONS 3

__u8 physmap_read8(struct map_info *map, unsigned long ofs)
{
    if(probe_done == 0)
        return __raw_readb(map->map_priv_1 + ofs);

    else if(ofs < DUPLCATED_BOOT_PART_PHY_START_OFFSET)
    {
        return __raw_readb(map->map_priv_1 + ofs);
    }
    else
    {
        if(ofs < (WINDOW_SIZE - DUPLICATED_BOOTL_PART_SIZE))
            return __raw_readb(map->map_priv_1 + ofs + DUPLICATED_BOOTL_PART_SIZE);
        else
            return __raw_readb(map->map_priv_1 + ofs - (DUPLCATED_BOOT_PART_PHY_START_OFFSET - DUPLICATED_BOOTL_PART_SIZE));
    }
}

__u16 physmap_read16(struct map_info *map, unsigned long ofs)
{
    if(probe_done == 0)
	    return __raw_readw(map->map_priv_1 + ofs);

    else if(ofs < DUPLCATED_BOOT_PART_PHY_START_OFFSET)
    {
        return __raw_readw(map->map_priv_1 + ofs);
    }
    else
    {
        if(ofs < (WINDOW_SIZE - DUPLICATED_BOOTL_PART_SIZE))
            return __raw_readw(map->map_priv_1 + ofs + DUPLICATED_BOOTL_PART_SIZE);
        else
            return __raw_readw(map->map_priv_1 + ofs - (DUPLCATED_BOOT_PART_PHY_START_OFFSET - DUPLICATED_BOOTL_PART_SIZE));
    }
}

__u32 physmap_read32(struct map_info *map, unsigned long ofs)
{
    if(probe_done == 0)
    	return __raw_readl(map->map_priv_1 + ofs);

    else if(ofs < DUPLCATED_BOOT_PART_PHY_START_OFFSET)
    {
        return __raw_readl(map->map_priv_1 + ofs);
    }
    else
    {
        if(ofs < (WINDOW_SIZE - DUPLICATED_BOOTL_PART_SIZE))
            return __raw_readl(map->map_priv_1 + ofs + DUPLICATED_BOOTL_PART_SIZE);
        else
            return __raw_readl(map->map_priv_1 + ofs - (DUPLCATED_BOOT_PART_PHY_START_OFFSET - DUPLICATED_BOOTL_PART_SIZE));
    }

}

void physmap_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
    if(probe_done == 0)
        memcpy_fromio(to, map->map_priv_1 + from, len);
    else if(from < DUPLCATED_BOOT_PART_PHY_START_OFFSET)
    {
        memcpy_fromio(to, map->map_priv_1 + from, len);
    }
    else
    {
        if(from < (WINDOW_SIZE - DUPLICATED_BOOTL_PART_SIZE))
            memcpy_fromio(to, map->map_priv_1 + from + DUPLICATED_BOOTL_PART_SIZE, len);
        else
            memcpy_fromio(to, map->map_priv_1 + from - (DUPLCATED_BOOT_PART_PHY_START_OFFSET - DUPLICATED_BOOTL_PART_SIZE), len);
    }
}

void physmap_write8(struct map_info *map, __u8 d, unsigned long adr)
{
    if(probe_done == 0)
    {
        __raw_writeb(d, map->map_priv_1 + adr);
    }
    else if(adr < DUPLCATED_BOOT_PART_PHY_START_OFFSET)
    {
        __raw_writeb(d, map->map_priv_1 + adr);
    }
    else
    {
        if(adr < (WINDOW_SIZE - DUPLICATED_BOOTL_PART_SIZE))
            __raw_writeb(d, map->map_priv_1 + adr + DUPLICATED_BOOTL_PART_SIZE);
        else
            __raw_writeb(d, map->map_priv_1 + adr - (DUPLCATED_BOOT_PART_PHY_START_OFFSET - DUPLICATED_BOOTL_PART_SIZE));
    }

    mb();
}


void physmap_write16(struct map_info *map, __u16 d, unsigned long adr)
{
   if(probe_done == 0)
    {
        __raw_writew(d, map->map_priv_1 + adr);
    }
    else if(adr < DUPLCATED_BOOT_PART_PHY_START_OFFSET)
    {
        __raw_writew(d, map->map_priv_1 + adr);
    }
    else
    {
        if(adr < (WINDOW_SIZE - DUPLICATED_BOOTL_PART_SIZE))
            __raw_writew(d, map->map_priv_1 + adr + DUPLICATED_BOOTL_PART_SIZE);
        else
            __raw_writew(d, map->map_priv_1 + adr - (DUPLCATED_BOOT_PART_PHY_START_OFFSET - DUPLICATED_BOOTL_PART_SIZE));
    }

    mb();

}

void physmap_write32(struct map_info *map, __u32 d, unsigned long adr)
{
   if(probe_done == 0)
        __raw_writel(d, map->map_priv_1 + adr);
    else if(adr < DUPLCATED_BOOT_PART_PHY_START_OFFSET)
        __raw_writel(d, map->map_priv_1 + adr);
    else
    {
        if(adr < (WINDOW_SIZE - DUPLICATED_BOOTL_PART_SIZE))
            __raw_writel(d, map->map_priv_1 + adr + DUPLICATED_BOOTL_PART_SIZE);
        else
            __raw_writel(d, map->map_priv_1 + adr - (DUPLCATED_BOOT_PART_PHY_START_OFFSET - DUPLICATED_BOOTL_PART_SIZE));
    }

    mb();

}

void physmap_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
    if(probe_done == 0)
    {
        memcpy_toio(map->map_priv_1 + to, from, len);
    }
    else if(to < DUPLCATED_BOOT_PART_PHY_START_OFFSET)
    {
        memcpy_toio(map->map_priv_1 + to, from, len);
    }
    else
    {
        if(to < (WINDOW_SIZE - DUPLICATED_BOOTL_PART_SIZE))
            memcpy_toio(map->map_priv_1 + to + DUPLICATED_BOOTL_PART_SIZE, from, len);
        else
            memcpy_toio(map->map_priv_1 + to - (DUPLCATED_BOOT_PART_PHY_START_OFFSET - DUPLICATED_BOOTL_PART_SIZE), from, len);
    }

}

struct map_info physmap_map = {
	name: "Physically mapped flash",
	size: WINDOW_SIZE,
	buswidth: BUSWIDTH,
	read8: physmap_read8,
	read16: physmap_read16,
	read32: physmap_read32,
	copy_from: physmap_copy_from,
	write8: physmap_write8,
	write16: physmap_write16,
	write32: physmap_write32,
	copy_to: physmap_copy_to
};

#if LINUX_VERSION_CODE < 0x20212 && defined(MODULE)
#define init_physmap init_module
#define cleanup_physmap cleanup_module
#endif

mod_init_t init_physmap(void)
{
    printk(KERN_NOTICE "physmap flash device: %x at %x\n", WINDOW_SIZE, WINDOW_ADDR);
//	physmap_map.map_priv_1 = (unsigned long)ioremap(WINDOW_ADDR, WINDOW_SIZE);
	physmap_map.map_priv_1 = (unsigned long)(WINDOW_ADDR);

	if (!physmap_map.map_priv_1) {
		printk("Failed to ioremap\n");
		return -EIO;
	}
	mymtd = do_cfi_probe(&physmap_map);
    probe_done = 1;
#ifdef TEST_FLASH
    printk("\n probe done no of part = %d\n", NUM_PARTITIONS);
    {
        int i;
        for(i = 0; i < NUM_PARTITIONS; i++)
        {
            printk("name  : %s\n", myboard_partitions[i].name);
            printk("offset: %x\n", myboard_partitions[i].offset);
            printk("size  : %x\n", myboard_partitions[i].size);
        }
    }
#endif
	if (mymtd) {
#ifdef MODULE
		mymtd->module = &__this_module;
#endif
        /* Create MTD devices for each partition. */
        add_mtd_partitions( mymtd, myboard_partitions, NUM_PARTITIONS );

		return 0;
	}

//	iounmap((void *)physmap_map.map_priv_1);
	return -ENXIO;
}

mod_exit_t cleanup_physmap(void)
{
	if (mymtd) {
		del_mtd_device(mymtd);
		map_destroy(mymtd);
	}
	if (physmap_map.map_priv_1) {
//		iounmap((void *)physmap_map.map_priv_1);
		physmap_map.map_priv_1 = 0;
	}
}

module_init(init_physmap);
module_exit(cleanup_physmap);

