/*
 *  linux/fs/proc/proc_misc.c
 *
 *  linux/fs/proc/array.c
 *  Copyright (C) 1992  by Linus Torvalds
 *  based on ideas by Darren Senn
 *
 *  This used to be the part of array.c. See the rest of history and credits
 *  there. I took this into a separate file and switched the thing to generic
 *  proc_file_inode_operations, leaving in array.c only per-process stuff.
 *  Inumbers allocation made dynamic (via create_proc_entry()).  AV, May 1999.
 *
 * Changes:
 * Fulton Green      :  Encapsulated position metric calculations.
 *			<kernel@FultonGreen.com>
 */

#include <linux/types.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/kernel_stat.h>
#include <linux/tty.h>
#include <linux/string.h>
#include <linux/mman.h>
#include <linux/proc_fs.h>
#include <linux/ioport.h>
#include <linux/config.h>
#include <linux/mm.h>
#include <linux/pagemap.h>
#include <linux/swap.h>
#include <linux/slab.h>
#include <linux/smp.h>
#include <linux/signal.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/smp_lock.h>

#ifdef ISHONI_ENGINE
#include <linux/mmzone.h>
#include <linux/swapctl.h>
#endif

#include <asm/uaccess.h>
#include <asm/pgtable.h>
#include <asm/io.h>


#define LOAD_INT(x) ((x) >> FSHIFT)
#define LOAD_FRAC(x) LOAD_INT(((x) & (FIXED_1-1)) * 100)
/*
 * Warning: stuff below (imported functions) assumes that its output will fit
 * into one page. For some of those functions it may be wrong. Moreover, we
 * have a way to deal with that gracefully. Right now I used straightforward
 * wrappers, but this needs further analysis wrt potential overflows.
 */
extern int get_cpuinfo(char *);
extern int get_hardware_list(char *);
extern int get_stram_list(char *);
#ifdef CONFIG_DEBUG_MALLOC
extern int get_malloc(char * buffer);
#endif
#ifdef CONFIG_MODULES
extern int get_module_list(char *);
extern int get_ksyms_list(char *, char **, off_t, int);
#endif
extern int get_device_list(char *);
extern int get_partition_list(char *, char **, off_t, int);
extern int get_filesystem_list(char *);
extern int get_filesystem_info(char *);
extern int get_exec_domain_list(char *);
extern int get_irq_list(char *);
extern int get_dma_list(char *);
extern int get_locks_status (char *, char **, off_t, int);
extern int get_swaparea_info (char *);
#ifdef CONFIG_SGI_DS1286
extern int get_ds1286_status(char *);
#endif

static int proc_calc_metrics(char *page, char **start, off_t off,
				 int count, int *eof, int len)
{
	if (len <= off+count) *eof = 1;
	*start = page + off;
	len -= off;
	if (len>count) len = count;
	if (len<0) len = 0;
	return len;
}

static int loadavg_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int a, b, c;
	int len;

	a = avenrun[0] + (FIXED_1/200);
	b = avenrun[1] + (FIXED_1/200);
	c = avenrun[2] + (FIXED_1/200);
	len = sprintf(page,"%d.%02d %d.%02d %d.%02d %d/%d %d\n",
		LOAD_INT(a), LOAD_FRAC(a),
		LOAD_INT(b), LOAD_FRAC(b),
		LOAD_INT(c), LOAD_FRAC(c),
		nr_running, nr_threads, last_pid);
	return proc_calc_metrics(page, start, off, count, eof, len);
}

static int uptime_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	unsigned long uptime;
	unsigned long idle;
	int len;

	uptime = jiffies;
	idle = init_tasks[0]->times.tms_utime + init_tasks[0]->times.tms_stime;

	/* The formula for the fraction parts really is ((t * 100) / HZ) % 100, but
	   that would overflow about every five days at HZ == 100.
	   Therefore the identity a = (a / b) * b + a % b is used so that it is
	   calculated as (((t / HZ) * 100) + ((t % HZ) * 100) / HZ) % 100.
	   The part in front of the '+' always evaluates as 0 (mod 100). All divisions
	   in the above formulas are truncating. For HZ being a power of 10, the
	   calculations simplify to the version in the #else part (if the printf
	   format is adapted to the same number of digits as zeroes in HZ.
	 */
#if HZ!=100
	len = sprintf(page,"%lu.%02lu %lu.%02lu\n",
		uptime / HZ,
		(((uptime % HZ) * 100) / HZ) % 100,
		idle / HZ,
		(((idle % HZ) * 100) / HZ) % 100);
#else
	len = sprintf(page,"%lu.%02lu %lu.%02lu\n",
		uptime / HZ,
		uptime % HZ,
		idle / HZ,
		idle % HZ);
#endif
	return proc_calc_metrics(page, start, off, count, eof, len);
}

#ifdef ISHONI_ENGINE

struct page_stats {
	unsigned int vmlen;
	unsigned int lockedlen;
	unsigned int rss;
	unsigned int reserved;
	unsigned int locked;
	unsigned int accessed;
	unsigned int clean;
	unsigned int dirty;
	unsigned int dirty_mapped;
	unsigned int dirty_anonymous;
};

static void get_page_stats (struct vm_area_struct* vma, pte_t pte, struct page_stats *stats)
{
	struct page *page = pte_page(pte);

	if (!VALID_PAGE(page)) {
		printk ("get_page_stats: Invalid page\n");
		return;
	}

	stats->rss++;

	if (PageReserved(page)) {
		stats->reserved++;
		return;
	}
	if (pte_young(pte)) {
		stats->accessed++;
		return;
	}
	if (PageLocked(page)) {
		stats->locked++;
		return;
	}
	if (!pte_dirty(pte)) {
		if (page->age)
			stats->accessed++;
		else stats->clean++;
		return;
	}

#if 0
	if (vma->vm_flags & VM_EXEC) {
		if (vma->vm_flags & VM_EXECUTABLE) {
			printk ("Dirty executable page! vma flags = %x\n", vma->vm_flags);
		}
	}
#endif

	if (page->mapping) {
		stats->dirty_mapped++;
		stats->dirty++;
		return;
	}
	stats->dirty_anonymous++;
	stats->dirty++;
}

static void vma_stats (struct vm_area_struct * vma, struct page_stats *stats)
{
	unsigned long address;
	pgd_t *pgd;
	pmd_t *pmd;

	if (vma->vm_flags & (VM_LOCKED|VM_RESERVED)) {
		stats->lockedlen += (vma->vm_end - vma->vm_start) >> 10;
		return;
	}

	stats->vmlen += (vma->vm_end - vma->vm_start) >> 10;
	for (address = vma->vm_start; address < vma->vm_end; address += PAGE_SIZE)
	{
		pgd = pgd_offset(vma->vm_mm, address);
		pmd = pmd_offset(pgd, address);

		if (pmd) {
			pte_t *pte = pte_offset(pmd, address);
			if (pte && pte_present(*pte))
				get_page_stats (vma, *pte, stats);
		}
	}
}

/* 
 * If buf is non-null prints formatted output into
 * the buf else prints on the console. This function
 * is useful when we want the same code to print on
 * the console as well as in the proc files.
 */

static int my_printf (char *buf, const char *fmt, ...)
{
	va_list args;
	int len;
	static char lbuf[1024];

	va_start(args, fmt);
	if (buf) {
		len = vsprintf (buf, fmt, args);
		va_end(args);
		return len;
	}
	vsprintf (lbuf, fmt, args);
	va_end(args);
	printk ("%s", lbuf);
	return 0;
}

int show_mem_status (char *page)
{
	unsigned long order;
	unsigned type;
	int len;

	len = my_printf(page, "(Active: %d, inactive_dirty: %d, inactive_clean: %d, free: %d (%d %d %d) )\n",
		      nr_active_pages,
		      nr_inactive_dirty_pages,
		      nr_inactive_clean_pages(),
		      nr_free_pages(),
		      freepages.min,
		      freepages.low,
		      freepages.high);

	for (type = 0; type < MAX_NR_ZONES; type++) {
		struct list_head *head, *curr;
		zone_t *zone = pgdat_list->node_zones + type;
 		unsigned long nr, total, flags;

		total = 0;
		if (zone->size) {
			len += my_printf (page + len, "zone %s(", zone->name);
			spin_lock_irqsave(&zone->lock, flags);
		 	for (order = 0; order < MAX_ORDER; order++) {
				head = &(zone->free_area + order)->free_list;
				curr = head;
				nr = 0;
				for (;;) {
					curr = curr->next;
					if (curr == head)
						break;
					nr++;
				}
				total += nr * (1 << order);
				len += my_printf(page + len, "%lu*%lukB ", nr,
						(PAGE_SIZE>>10) << order);
			}
			spin_unlock_irqrestore(&zone->lock, flags);
			len += my_printf(page + len, "= %lukB)\n", total * (PAGE_SIZE>>10));
		}
	}
	return len;
}

#define MEMUSAGE_HEADER_FORMAT	"%3s (%13s)%5s%13s%13s%17s%17s%13s\n"
#define MEMUSAGE_ROW_FORMAT	"%3d (%13s)%5lu%4lu(%3lu,%3lu)%4lu(%3lu,%3lu)%4lu(%3lu,%3lu,%3lu)%4lu(%3lu,%3lu,%3lu)%4lu(%3lu,%3lu)\n"
#define MEMUSAGE_ROW_LEN		100

static int __show_task_mem (char *page, int can_sleep)
{
	struct task_struct *p = NULL;
	int len;

	len = my_printf (page, MEMUSAGE_HEADER_FORMAT,
			"PID", "Task Name", 
			"rss", "dat(a,d)", "stk(a,d)",
			"exe(a,c,d)", "lib(a,c,d)", "wrt(a,d)");

	read_lock(&tasklist_lock);
	for_each_task(p) {

		struct vm_area_struct * vma;
		struct page_stats data_stats, stack_stats;
		struct page_stats exec_stats, lib_stats;
		struct page_stats wrt_stats;

		memset (&data_stats, 0, sizeof (struct page_stats));
		memset (&stack_stats, 0, sizeof (struct page_stats));
		memset (&exec_stats, 0, sizeof (struct page_stats));
		memset (&lib_stats, 0, sizeof (struct page_stats));
		memset (&wrt_stats, 0, sizeof (struct page_stats));

		if (p->pid && p->mm) {
			struct mm_struct *mm = p->mm;
			int gotlock;

			if (can_sleep) {
				down (&mm->mmap_sem);
				gotlock = 1;
			}
			else gotlock = !down_trylock(&mm->mmap_sem);

			if (gotlock) {

				for (vma = mm->mmap; vma; vma = vma->vm_next) {
					if (!vma->vm_file) {
						if (vma->vm_flags & VM_GROWSDOWN)
							vma_stats (vma, &stack_stats);
						else vma_stats (vma, &data_stats);
						continue;
					}
					if (vma->vm_flags & VM_WRITE) {
						vma_stats (vma, &wrt_stats);
						continue;
					}
					if (vma->vm_flags & VM_EXEC) {
						if (vma->vm_flags & VM_EXECUTABLE)
							vma_stats(vma, &exec_stats);
						else vma_stats(vma, &lib_stats);
						continue;
					}
				}

				up(&mm->mmap_sem);
			}

			/* Don't write beyond one page boundary */
			if (page && ((page + len + MEMUSAGE_ROW_LEN) >= (char *)PAGE_ALIGN((unsigned long)page)))
			{
				printk ("show_task_mem: too many tasks to fit in a single page\n");
				break;
			}

			len += my_printf(page + len, MEMUSAGE_ROW_FORMAT,
				       p->pid, p->comm, mm->rss,

					data_stats.rss, data_stats.accessed, data_stats.dirty,
					stack_stats.rss, stack_stats.accessed, stack_stats.dirty,
					exec_stats.rss, exec_stats.accessed, exec_stats.clean, exec_stats.dirty,
					lib_stats.rss, lib_stats.accessed, lib_stats.clean, lib_stats.dirty,
					wrt_stats.rss, wrt_stats.accessed, wrt_stats.dirty);
		}
	}
	read_unlock(&tasklist_lock);
	return len;
}

int show_task_mem (char *page)
{
	return __show_task_mem (page, 0);
}

static int memusage_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len;

	len = show_mem_status (page);
	len += __show_task_mem (page + len, 1);
	return proc_calc_metrics(page, start, off, count, eof, len);
}

int show_thrash_info (char *page)
{
	struct task_struct *p = NULL;
	unsigned long total_major = 0, total_minor = 0;
	unsigned long total_n_cachehit_exec = 0, total_n_fmap_nopage_exec = 0;
	unsigned long total_n_cramfs_nopage_exec = 0;
	int len;

	len = my_printf (page, "%3s %13s %13s %13s %13s %13s %13s\n",
		       	"PID", "Task Name", "Major faults", "Minor faults",
		       	"Fmap nopage", "cache hit", "readpage");

	read_lock(&tasklist_lock);
	for_each_task(p) {
		if (p->pid && p->mm) {
			len += my_printf (page + len, "%3d %13s %13lu %13lu %13lu %13lu %13lu\n",
				       p->pid, p->comm, p->maj_flt, p->min_flt,
				       p->n_fmap_nopage_exec, p->n_cachehit_exec, p->n_cramfs_nopage_exec);
			total_major += p->maj_flt;
			total_minor += p->min_flt;
			total_n_cachehit_exec += p->n_cachehit_exec;
			total_n_fmap_nopage_exec += p->n_fmap_nopage_exec;
			total_n_cramfs_nopage_exec += p->n_cramfs_nopage_exec;
		}
	}
	len += my_printf (page + len, "%3d %13s %13lu %13lu %13lu %13lu %13lu\n",
		       	0, "TOTAL", total_major, total_minor,
		       	total_n_fmap_nopage_exec, total_n_cachehit_exec, total_n_cramfs_nopage_exec);
	read_unlock(&tasklist_lock);
	return len;
}

static int thrash_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len;

	len = show_thrash_info (page);
	return proc_calc_metrics(page, start, off, count, eof, len);
}
#endif /* ISHONI_ENGINE */

static int meminfo_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	struct sysinfo i;
	int len;

/*
 * display in kilobytes.
 */
#define K(x) ((x) << (PAGE_SHIFT - 10))
#define B(x) ((x) << PAGE_SHIFT)
        si_meminfo(&i);
        si_swapinfo(&i);
        len = sprintf(page, "        total:    used:    free:  shared: buffers:  cached:\n"
                "Mem:  %8lu %8lu %8lu %8lu %8lu %8u\n"
                "Swap: %8lu %8lu %8lu\n",
                B(i.totalram), B(i.totalram-i.freeram), B(i.freeram),
                B(i.sharedram), B(i.bufferram),
                B(atomic_read(&page_cache_size)), B(i.totalswap),
                B(i.totalswap-i.freeswap), B(i.freeswap));
        /*
         * Tagged format, for easy grepping and expansion.
         * The above will go away eventually, once the tools
         * have been updated.
         */
        len += sprintf(page+len,
                "MemTotal:     %8lu kB\n"
                "MemFree:      %8lu kB\n"
                "MemShared:    %8lu kB\n"
                "Buffers:      %8lu kB\n"
                "Cached:       %8u kB\n"
		"Active:       %8u kB\n"
		"Inact_dirty:  %8u kB\n"
		"Inact_clean:  %8u kB\n"
		"Inact_target: %8lu kB\n"
                "HighTotal:    %8lu kB\n"
                "HighFree:     %8lu kB\n"
                "LowTotal:     %8lu kB\n"
                "LowFree:      %8lu kB\n"
                "SwapTotal:    %8lu kB\n"
                "SwapFree:     %8lu kB\n",
                K(i.totalram),
                K(i.freeram),
                K(i.sharedram),
                K(i.bufferram),
                K(atomic_read(&page_cache_size)),
		K(nr_active_pages),
		K(nr_inactive_dirty_pages),
		K(nr_inactive_clean_pages()),
		K(inactive_target),
                K(i.totalhigh),
                K(i.freehigh),
                K(i.totalram-i.totalhigh),
                K(i.freeram-i.freehigh),
                K(i.totalswap),
                K(i.freeswap));

	return proc_calc_metrics(page, start, off, count, eof, len);
#undef B
#undef K
}

static int version_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	extern char *linux_banner;
	int len;

	strcpy(page, linux_banner);
	len = strlen(page);
	return proc_calc_metrics(page, start, off, count, eof, len);
}

static int cpuinfo_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len = get_cpuinfo(page);
	return proc_calc_metrics(page, start, off, count, eof, len);
}

#ifdef CONFIG_PROC_HARDWARE
static int hardware_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len = get_hardware_list(page);
	return proc_calc_metrics(page, start, off, count, eof, len);
}
#endif

#ifdef CONFIG_STRAM_PROC
static int stram_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len = get_stram_list(page);
	return proc_calc_metrics(page, start, off, count, eof, len);
}
#endif

#ifdef CONFIG_DEBUG_MALLOC
static int malloc_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len = get_malloc(page);
	return proc_calc_metrics(page, start, off, count, eof, len);
}
#endif

#ifdef CONFIG_MODULES
static int modules_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len = get_module_list(page);
	return proc_calc_metrics(page, start, off, count, eof, len);
}

static int ksyms_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len = get_ksyms_list(page, start, off, count);
	if (len < count) *eof = 1;
	return len;
}
#endif

static int kstat_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int i, len;
	extern unsigned long total_forks;
	unsigned long jif = jiffies;
	unsigned int sum = 0, user = 0, nice = 0, system = 0;
	int major, disk;

	for (i = 0 ; i < smp_num_cpus; i++) {
		int cpu = cpu_logical_map(i), j;

		user += kstat.per_cpu_user[cpu];
		nice += kstat.per_cpu_nice[cpu];
		system += kstat.per_cpu_system[cpu];
		for (j = 0 ; j < NR_IRQS ; j++)
			sum += kstat.irqs[cpu][j];
	}

	len = sprintf(page, "cpu  %u %u %u %lu\n", user, nice, system,
		      jif * smp_num_cpus - (user + nice + system));
	for (i = 0 ; i < smp_num_cpus; i++)
		len += sprintf(page + len, "cpu%d %u %u %u %lu\n",
			i,
			kstat.per_cpu_user[cpu_logical_map(i)],
			kstat.per_cpu_nice[cpu_logical_map(i)],
			kstat.per_cpu_system[cpu_logical_map(i)],
			jif - (  kstat.per_cpu_user[cpu_logical_map(i)] \
			           + kstat.per_cpu_nice[cpu_logical_map(i)] \
			           + kstat.per_cpu_system[cpu_logical_map(i)]));
	len += sprintf(page + len,
		"page %u %u\n"
                "swap %u %u\n"
		"intr %u",
			kstat.pgpgin,
			kstat.pgpgout,
			kstat.pswpin,
			kstat.pswpout,
			sum
	);
	for (i = 0 ; i < NR_IRQS ; i++)
		len += sprintf(page + len, " %u", kstat_irqs(i));

	len += sprintf(page + len, "\ndisk_io: ");

	for (major = 0; major < DK_MAX_MAJOR; major++) {
		for (disk = 0; disk < DK_MAX_DISK; disk++) {
			int active = kstat.dk_drive[major][disk] +
				kstat.dk_drive_rblk[major][disk] +
				kstat.dk_drive_wblk[major][disk];
			if (active)
				len += sprintf(page + len,
					"(%u,%u):(%u,%u,%u,%u,%u) ",
					major, disk,
					kstat.dk_drive[major][disk],
					kstat.dk_drive_rio[major][disk],
					kstat.dk_drive_rblk[major][disk],
					kstat.dk_drive_wio[major][disk],
					kstat.dk_drive_wblk[major][disk]
			);
		}
	}

	len += sprintf(page + len,
		"\nctxt %u\n"
		"btime %lu\n"
		"processes %lu\n",
		kstat.context_swtch,
		xtime.tv_sec - jif / HZ,
		total_forks);

	return proc_calc_metrics(page, start, off, count, eof, len);
}

static int devices_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len = get_device_list(page);
	return proc_calc_metrics(page, start, off, count, eof, len);
}

static int partitions_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len = get_partition_list(page, start, off, count);
	if (len < count) *eof = 1;
	return len;
}

#if !defined(CONFIG_ARCH_S390)
static int interrupts_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len = get_irq_list(page);
	return proc_calc_metrics(page, start, off, count, eof, len);
}
#endif

static int filesystems_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len = get_filesystem_list(page);
	return proc_calc_metrics(page, start, off, count, eof, len);
}

static int dma_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len = get_dma_list(page);
	return proc_calc_metrics(page, start, off, count, eof, len);
}

static int ioports_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len = get_ioport_list(page);
	return proc_calc_metrics(page, start, off, count, eof, len);
}

static int cmdline_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	extern char saved_command_line[];
	int len;

	len = sprintf(page, "%s\n", saved_command_line);
	len = strlen(page);
	return proc_calc_metrics(page, start, off, count, eof, len);
}

#ifdef CONFIG_SGI_DS1286
static int ds1286_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len = get_ds1286_status(page);
	return proc_calc_metrics(page, start, off, count, eof, len);
}
#endif

static int locks_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len;
	lock_kernel();
	len = get_locks_status(page, start, off, count);
	unlock_kernel();
	if (len < count) *eof = 1;
	return len;
}

static int mounts_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len = get_filesystem_info(page);
	return proc_calc_metrics(page, start, off, count, eof, len);
}

static int execdomains_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len = get_exec_domain_list(page);
	return proc_calc_metrics(page, start, off, count, eof, len);
}

static int swaps_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len = get_swaparea_info(page);
	return proc_calc_metrics(page, start, off, count, eof, len);
}

static int memory_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len = get_mem_list(page);
	return proc_calc_metrics(page, start, off, count, eof, len);
}

/*
 * This function accesses profiling information. The returned data is
 * binary: the sampling step and the actual contents of the profile
 * buffer. Use of the program readprofile is recommended in order to
 * get meaningful info out of these data.
 */
static ssize_t read_profile(struct file *file, char *buf,
			    size_t count, loff_t *ppos)
{
	unsigned long p = *ppos;
	ssize_t read;
	char * pnt;
	unsigned int sample_step = 1 << prof_shift;

	if (p >= (prof_len+1)*sizeof(unsigned int))
		return 0;
	if (count > (prof_len+1)*sizeof(unsigned int) - p)
		count = (prof_len+1)*sizeof(unsigned int) - p;
	read = 0;

	while (p < sizeof(unsigned int) && count > 0) {
		put_user(*((char *)(&sample_step)+p),buf);
		buf++; p++; count--; read++;
	}
	pnt = (char *)prof_buffer + p - sizeof(unsigned int);
	copy_to_user(buf,(void *)pnt,count);
	read += count;
	*ppos += read;
	return read;
}

/*
 * Writing to /proc/profile resets the counters
 *
 * Writing a 'profiling multiplier' value into it also re-sets the profiling
 * interrupt frequency, on architectures that support this.
 */
static ssize_t write_profile(struct file * file, const char * buf,
			     size_t count, loff_t *ppos)
{
#ifdef CONFIG_SMP
	extern int setup_profiling_timer (unsigned int multiplier);

	if (count==sizeof(int)) {
		unsigned int multiplier;

		if (copy_from_user(&multiplier, buf, sizeof(int)))
			return -EFAULT;

		if (setup_profiling_timer(multiplier))
			return -EINVAL;
	}
#endif

	memset(prof_buffer, 0, prof_len * sizeof(*prof_buffer));
	return count;
}

static struct file_operations proc_profile_operations = {
	read:		read_profile,
	write:		write_profile,
};

struct proc_dir_entry *proc_root_kcore;

void __init proc_misc_init(void)
{
	struct proc_dir_entry *entry;
	static struct {
		char *name;
		int (*read_proc)(char*,char**,off_t,int,int*,void*);
	} *p, simple_ones[] = {
		{"loadavg",     loadavg_read_proc},
		{"uptime",	uptime_read_proc},
		{"meminfo",	meminfo_read_proc},
		{"version",	version_read_proc},
		{"cpuinfo",	cpuinfo_read_proc},
#ifdef ISHONI_ENGINE
		{"memusage",	memusage_read_proc},
		{"thrash",	thrash_read_proc},
#endif /* ISHONI_ENGINE */
#ifdef CONFIG_PROC_HARDWARE
		{"hardware",	hardware_read_proc},
#endif
#ifdef CONFIG_STRAM_PROC
		{"stram",	stram_read_proc},
#endif
#ifdef CONFIG_DEBUG_MALLOC
		{"malloc",	malloc_read_proc},
#endif
#ifdef CONFIG_MODULES
		{"modules",	modules_read_proc},
		{"ksyms",	ksyms_read_proc},
#endif
		{"stat",	kstat_read_proc},
		{"devices",	devices_read_proc},
		{"partitions",	partitions_read_proc},
#if !defined(CONFIG_ARCH_S390)
		{"interrupts",	interrupts_read_proc},
#endif
		{"filesystems",	filesystems_read_proc},
		{"dma",		dma_read_proc},
		{"ioports",	ioports_read_proc},
		{"cmdline",	cmdline_read_proc},
#ifdef CONFIG_SGI_DS1286
		{"rtc",		ds1286_read_proc},
#endif
		{"locks",	locks_read_proc},
		{"mounts",	mounts_read_proc},
		{"swaps",	swaps_read_proc},
		{"iomem",	memory_read_proc},
		{"execdomains",	execdomains_read_proc},
		{NULL,}
	};
	for (p = simple_ones; p->name; p++)
		create_proc_read_entry(p->name, 0, NULL, p->read_proc, NULL);

	/* And now for trickier ones */
	entry = create_proc_entry("kmsg", S_IRUSR, &proc_root);
	if (entry)
		entry->proc_fops = &proc_kmsg_operations;
	proc_root_kcore = create_proc_entry("kcore", S_IRUSR, NULL);
	if (proc_root_kcore) {
		proc_root_kcore->proc_fops = &proc_kcore_operations;
		proc_root_kcore->size =
				(size_t)high_memory - PAGE_OFFSET + PAGE_SIZE;
	}
	if (prof_shift) {
		entry = create_proc_entry("profile", S_IWUSR | S_IRUGO, NULL);
		if (entry) {
			entry->proc_fops = &proc_profile_operations;
			entry->size = (1+prof_len) * sizeof(unsigned int);
		}
	}
#ifdef __powerpc__
	{
		extern struct file_operations ppc_htab_operations;
		entry = create_proc_entry("ppc_htab", S_IRUGO|S_IWUSR, NULL);
		if (entry)
			entry->proc_fops = &ppc_htab_operations;
	}
#endif
	entry = create_proc_read_entry("slabinfo", S_IWUSR | S_IRUGO, NULL,
				       slabinfo_read_proc, NULL);
	if (entry)
		entry->write_proc = slabinfo_write_proc;
}
